import os
import stat
import time
import urllib
from typing import IO, Dict, Iterable, Optional, Union

from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info
from ratarmountcore.utils import overrides

try:
    import fsspec
    import fsspec.core
except ImportError:
    fsspec = None  # type: ignore

try:
    from fsspec.implementations.http import HTTPFileSystem
except ImportError:
    HTTPFileSystem = None  # type: ignore

try:
    from fsspec.implementations.github import GithubFileSystem
except ImportError:
    GithubFileSystem = None  # type: ignore

try:
    from webdav4.fsspec import WebdavFileSystem
except ImportError:
    WebdavFileSystem = None  # type: ignore

try:
    from dropboxdrivefs import DropboxDriveFileSystem
except ImportError:
    DropboxDriveFileSystem = None  # type: ignore


class FSSpecMountSource(MountSource):
    """
    Generic wrapper around fsspec-based filesystems.
    At least as "generic" as it gets given that many details are inconsistent between the implementations.
    Note also that many implementations are rather experimental, utterly slow, or unstable.
    """

    # TODO implement some of the most rudimentarily implemented filesystems myself instead of via fsspec.
    # wc -l 'fsspec/implementations/'*.py | sort -n
    #      0 fsspec/implementations/__init__.py
    #     58 fsspec/implementations/data.py
    #     75 fsspec/implementations/cache_mapper.py
    #    124 fsspec/implementations/jupyter.py
    #    124 fsspec/implementations/tar.py              -> SQLiteIndexedTar
    #    127 fsspec/implementations/git.py              -> TODO
    #    152 fsspec/implementations/dask.py
    #    176 fsspec/implementations/zip.py              -> ZipMountSource
    #    180 fsspec/implementations/sftp.py             -> fsspec/sshfs
    #    213 fsspec/implementations/libarchive.py       -> LibarchiveMountSource
    #    232 fsspec/implementations/cache_metadata.py
    #    239 fsspec/implementations/github.py
    #    303 fsspec/implementations/memory.py
    #    304 fsspec/implementations/arrow.py
    #    372 fsspec/implementations/dirfs.py            -> FolderMountSource + chdir
    #    395 fsspec/implementations/ftp.py
    #    416 fsspec/implementations/smb.py
    #    467 fsspec/implementations/dbfs.py
    #    471 fsspec/implementations/local.py            -> FolderMountSource
    #    484 fsspec/implementations/webhdfs.py
    #    872 fsspec/implementations/http.py
    #    929 fsspec/implementations/cached.py
    #   1173 fsspec/implementations/reference.py
    # I guess git is the most obvious candidate because it is the most interesting and most barebone implementation.

    # pylint: disable=unused-argument
    def __init__(self, urlOrFS, prefix: Optional[str] = None, **options) -> None:
        """
        urlOrFS : Take a URL or an already opened fsspec Filesystem object.
                  Note that this might take an AbstractFileSystem-derived object in the future.
        """
        if isinstance(urlOrFS, fsspec.AbstractFileSystem):
            fs = urlOrFS
        elif isinstance(urlOrFS, str):
            url_to_fs = fsspec.url_to_fs if hasattr(fsspec, 'url_to_fs') else fsspec.core.url_to_fs
            fs, path = url_to_fs(urlOrFS)
            if prefix is None:
                prefix = path
        else:
            raise ValueError("First argument must be an URL or inherit from fsspec.AbstractFileSystem!")
        self.fileSystem: fsspec.AbstractFileSystem = fs
        self.rootFileInfo = create_root_file_info(userdata=["/"])

        # The fsspec filesystems are not uniform! http:// expects the arguments to isdir with prefixed
        # protocol while other filesystem implementations are fine with only the path.
        #  - https://github.com/ray-project/ray/issues/26423#issuecomment-1179561181
        #  - https://github.com/fsspec/filesystem_spec/issues/1713
        #  - https://github.com/skshetry/webdav4/issues/198
        self._pathsRequireQuoting = HTTPFileSystem is not None and isinstance(self.fileSystem, HTTPFileSystem)
        if WebdavFileSystem:
            self._pathsRequireQuoting = self._pathsRequireQuoting or isinstance(self.fileSystem, WebdavFileSystem)
        self.prefix = prefix.rstrip("/") if prefix and prefix.strip("/") and self.fileSystem.isdir(prefix) else None
        self._pathsWithoutLeadingSlash = GithubFileSystem is not None and isinstance(self.fileSystem, GithubFileSystem)

    def _get_path(self, path: str) -> str:
        if self._pathsRequireQuoting:
            path = urllib.parse.quote(path)
        if self.prefix:
            if not path or path == "/":
                return self.prefix
            return self.prefix.rstrip("/") + "/" + path.lstrip("/")
        if self._pathsWithoutLeadingSlash:
            return path.lstrip("/")
        # The fsspec TAR implementation is verbatim. The path must exactly match the one in the archive, even
        # if the archive path is not normalized. E.g., compare with these:
        #  - tests/nested-symlinks.tar -> leading /. Did not work because
        #  - tests/nested-tar.tar -> no leading / only works with this fix
        #  - tests/single-file-with-leading-dot-slash.tar -> not working at all.
        #       We cannot test all possible denormalized paths when a normalized path is requested via FUSE.
        if not self.fileSystem.lexists(path) and self.fileSystem.lexists(path.lstrip("/")):
            path = path.lstrip("/")
        return path

    @staticmethod
    def _get_mode(entry) -> int:
        return 0o555 | (stat.S_IFDIR if entry.get('type', '') == 'directory' else stat.S_IFREG)

    @staticmethod
    def _get_modification_time(entry) -> Union[int, float]:
        # There is no standardized API for the modification time:
        # https://github.com/fsspec/filesystem_spec/issues/1680#issuecomment-2368750882
        #
        # sshfs.SSHF: 'mtime': datetime.datetime(2020, 3, 23, 20, 15, 34)
        # fsspec.implementations.git.GitFileSystem: Nothing with listdir(details=True)!
        # fsspec.implementations.ftp.FTPFileSystem: 'modify': '20241004165129'
        mtime = entry.get('mtime', None)
        if mtime is not None:
            return mtime.timestamp() if hasattr(mtime, 'timestamp') else mtime
        modify = entry.get('modify', None)
        if isinstance(modify, str):
            return time.mktime(time.strptime(modify, "%Y%m%d%H%M%S"))
        return 0

    @staticmethod
    def _convert_to_file_info(entry, path) -> FileInfo:
        # TODO fsspec does not have an API to get symbolic link targets!
        #      They kinda work only like hardlinks.
        # https://github.com/fsspec/filesystem_spec/issues/1679
        # https://github.com/fsspec/filesystem_spec/issues/1680
        size = entry.get('size', 0)
        # fmt: off
        return FileInfo(
            size     = size or 0,
            mtime    = FSSpecMountSource._get_modification_time(entry),
            mode     = FSSpecMountSource._get_mode(entry),
            linkname = "",
            uid      = os.getuid(),
            gid      = os.getgid(),
            userdata = [path],
        )
        # fmt: on

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def exists(self, path: str) -> bool:
        return self.fileSystem.lexists(self._get_path(path))

    def _list(self, path: str, onlyMode: bool) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        path = self._get_path(path)

        if path == '/' and DropboxDriveFileSystem and isinstance(self.fileSystem, DropboxDriveFileSystem):
            # We need to work around this obnoxious error:
            # dropbox.exceptions.BadInputError: BadInputError(
            #   '12345', 'Error in call to API function "files/list_folder":
            #    request body: path: Specify the root folder as an empty string rather than as "/".')
            # On the other hand, all paths must start with / or else they will not be found...
            path = ""

        result = self.fileSystem.listdir(path, detail=True)
        if not result:
            return []
        if isinstance(result[0], str):
            return result

        # Examples for listdir return values:
        #
        # sshfs.SSHF: [
        #   {'size': 8, 'type': 'link', 'gid': 0, 'uid': 0, 'time': datetime.datetime(2024, 10, 3, 19, 32, 42),
        #    'mtime': datetime.datetime(2020, 3, 23, 20, 15, 34), 'permissions': 41471, 'name': '/sbin'},
        #   {'size': 4096, 'type': 'directory', 'gid': 0, 'uid': 0, 'time': datetime.datetime(2024, 9, 25, 19, 45, 31),
        #    'mtime': datetime.datetime(2023, 7, 22, 11, 32, 1), 'permissions': 16877, 'name': '/var'}
        #   {'size': 134217728, 'type': 'file', 'gid': 0, 'uid': 0, 'time': datetime.datetime(2024, 9, 25, 19, 45, 30),
        #    'mtime': datetime.datetime(2021, 6, 16, 19, 26, 38), 'permissions': 33188, 'name': '/swapfile'}
        # -> "name" contains the absolute path to each file (also tested with subfolders)!
        # fsspec.implementations.git.GitFileSystem: [
        #   {'type': 'file', 'name': '.gitattributes', 'hex': '2a396079050e5847b7c995642ed07a7c8591bde9',
        #    'mode': '100644', 'size': 363},
        #   {'type': 'directory', 'name': '.github', 'hex': 'c8ab28a6ded46c96fa33a96a9d6d0b53dfe815de',
        #    'mode': '40000', 'size': 0},
        #   [{'type': 'directory', 'name': '.github/workflows', 'hex': 'b1b9b9b0d1ca1210f823195238e8fe71829fae42',
        #     'mode': '40000', 'size': 0}]
        # -> "name" is absolute path but without leading slash
        # fsspec.implementations.ftp.FTPFileSystem: [
        #   {'modify': '20241004165129', 'perm': 'el', 'size': 0, 'type': 'directory',
        #    'unique': 'fd01ga9f7f6', 'name': '/.git'},
        #   {'modify': '20240602192724', 'perm': 'r', 'size': 363, 'type': 'file',
        #    'unique': 'fd01g2de4e2', 'name': '/.gitattributes'},
        # fsspec.implementations.http.HTTPFileSystem: [
        #   {'name': 'http://127.0.0.1:8000/?S=D', 'size': None, 'type': 'file'},
        #   {'name': 'http://127.0.0.1:8000/benchmarks/', 'size': None, 'type': 'directory'},
        #   {'name': 'http://127.0.0.1:8000/benchmark-sshfs-block_size.py', 'size': None, 'type': 'file'},
        # -> For some reason, the name always has to include the full URL for the request and result.
        # -> There are some HTTP server artifacts such as "?S=D", which are links for changing the sorting...
        prefixToStrip = path.lstrip('/')
        result = {
            (
                entry['name'].strip('/')[len(prefixToStrip) :].strip('/')
                if entry['name'].strip('/').startswith(prefixToStrip)
                else entry['name']
            ): (
                FSSpecMountSource._get_mode(entry)
                if onlyMode
                else FSSpecMountSource._convert_to_file_info(entry, entry['name'])
            )
            for entry in result
        }

        # For some dumb reason, only the TAR filesystem, returns '/' for some subfolders.
        # It happens with nested-symlinks.tar, probably because the subfolder as its own entry in the TAR:
        #     drwx------ user/user  0 2020-04-10 10:46 /foo/
        #     lrwxrwxrwx user/user  0 2020-04-10 10:46 /foo/ufo -> /foo/foo
        #     -rwx------ user/user  6 2020-04-09 17:59 /foo/foo
        #     drwx------ user/user  0 2020-04-09 17:57 /foo/fighter/
        #     lrwxrwxrwx user/user  0 2020-04-09 16:53 /foo/fighter/foo -> ../foo
        #     lrwxrwxrwx user/user  0 2020-04-09 16:52 /foo/fighter/python -> /usr/bin/python
        #     lrwxrwxrwx user/user  0 2020-04-09 18:01 /foo/iriya -> ../iriya
        if '' in result:
            del result['']

        # For HTTPFileSystem, we need to filter out the entries for sorting.
        # For WebDAV we do not even need to unquote! We get unquoted file names with ls!
        if isinstance(self.fileSystem, fsspec.implementations.http.HTTPFileSystem):
            return {
                urllib.parse.unquote(name): info for name, info in result.items() if not name.startswith(('?', '#'))
            }

        return result

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        return self._list(path, onlyMode=False)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        return self._list(path, onlyMode=True)

    def _lookup_http(self, path: str) -> Optional[FileInfo]:
        path = self._get_path(path)

        # Avoid aiohttp.client_exceptions.ClientResponseError: 404, message='Not Found'
        if not self.fileSystem.lexists(path):
            return None

        # fs.info will always return the given path to be file because it counts it as an HTML file ...
        # isdir works somewhat better, but it downloads the whole file!
        # https://github.com/fsspec/filesystem_spec/issues/1707
        # Therefore, only call it if the mimetype indicates an HTML file.
        # In the future it might be best to call listdir on the parent path to detect whether it is a folder or file.
        info = self.fileSystem.info(path)
        if info.get('mimetype', None) == 'text/html' and self.fileSystem.isdir(path):
            return FSSpecMountSource._convert_to_file_info({'type': 'directory'}, path)
        return FSSpecMountSource._convert_to_file_info(info, path)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        if isinstance(self.fileSystem, fsspec.implementations.http.HTTPFileSystem):
            return self._lookup_http(path)

        path = self._get_path(path)
        if path == '/' or not path:
            # We need to handle this specially because some filesystems, at least ssshfs.SSHFileSystem,
            # do not support 'info' on '/' and will cause an exception:
            #
            # Traceback (most recent call last):
            #   sshfs/utils.py", line 27, in wrapper
            #     return await func(*args, **kwargs)
            #   sshfs/spec.py", line 145, in _info
            #     attributes = await channel.stat(path)
            #   asyncssh/sftp.py", line 4616, in stat
            #     return await self._handler.stat(path, flags,
            #   asyncssh/sftp.py", line 2713, in stat
            #     return cast(SFTPAttrs,  await self._make_request(
            #   asyncssh/sftp.py", line 2468, in _make_request
            #     result = self._packet_handlers[resptype](self, resp)
            #   asyncssh/sftp.py", line 2484, in _process_status
            #     raise exc
            # asyncssh.sftp.SFTPNoSuchFile: No such file
            #
            # Dropbox also does not like this:
            #
            # dropbox.exceptions.BadInputError: BadInputError('12345',
            #   'Error in call to API function "files/get_metadata":
            #   request body: path: The root folder is unsupported.')
            return self.rootFileInfo.clone()

        if not self.fileSystem.lexists(path):
            return None
        return FSSpecMountSource._convert_to_file_info(self.fileSystem.info(path), path)

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return 1

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        path = fileInfo.userdata[-1]
        assert isinstance(path, str)
        return self.fileSystem.open(path, block_size=buffering if buffering >= 0 else None)

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        if hasattr(self.fileSystem, '__exit__'):
            self.fileSystem.__exit__(exception_type, exception_value, exception_traceback)



# pylint: disable=abstract-method,unused-argument

import stat

import fsspec

from .mountsource import MountSource
from .mountsource.formats.tar import SQLiteIndexedTar
from .utils import overrides


class MountSourceFileSystem(fsspec.spec.AbstractFileSystem):
    """A thin adaptor from the MountSource interface to the fsspec AbstractFileSystem interface."""

    cachable = False

    def __init__(self, mountSource: MountSource, **kwargs):
        super().__init__(**kwargs)
        self.mountSource = mountSource

    @classmethod
    def _stripProtocol(cls, path):  # noqa # Overwrites fsspec.spec.AbstractFileSystem._stripProtocol!
        return path[-len(cls.protocol) - 3] if path.startswith(cls.protocol + '://') else path

    @staticmethod
    def _file_info_to_dict(name, fileInfo):
        return {
            "type": "directory" if stat.S_ISDIR(fileInfo.mode) else "file",
            "name": name,
            "mode": f"{fileInfo.mode:o}",
            "size": fileInfo.size,
        }

    @overrides(fsspec.spec.AbstractFileSystem)
    def ls(self, path, detail=True, **kwargs):
        strippedPath = self._stripProtocol(path)
        if detail:
            result = self.mountSource.list(strippedPath)
            if result is None:
                raise FileNotFoundError(path)
            if not isinstance(result, dict):
                result = {name: self.mountSource.lookup(name) for name in result}
            return [self._file_info_to_dict(name, info) for name, info in result.items() if info is not None]

        result = self.mountSource.list_mode(strippedPath)
        if result is None:
            raise FileNotFoundError(path)
        return list(result.keys()) if isinstance(result, dict) else result

    @overrides(fsspec.spec.AbstractFileSystem)
    def info(self, path, **kwargs):
        result = self.mountSource.lookup(self._stripProtocol(path))
        if result is None:
            raise FileNotFoundError(path)
        return self._file_info_to_dict(path, result)

    @overrides(fsspec.spec.AbstractFileSystem)
    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_options=None,
        **kwargs,
    ):
        if mode != "rb":
            raise ValueError("Only binary reading is supported!")
        fileInfo = self.mountSource.lookup(self._stripProtocol(path))
        if fileInfo is None:
            raise FileNotFoundError(path)
        return self.mountSource.open(fileInfo, buffering=block_size or -1)


class SQLiteIndexedTarFileSystem(MountSourceFileSystem):
    """
    Browse the files of a (compressed) TAR archive quickly.

    This is a more optimized alternative to fsspec.implementations.TarFileSystem.
    """

    protocol = "ratar"

    def __init__(
        self,
        # It must be called "fo" for URL chaining to work!
        # https://filesystem-spec.readthedocs.io/en/latest/features.html#url-chaining
        fo=None,
        *,  # force all parameters after to be keyword-only
        target_options=None,
        target_protocol=None,
        **kwargs,
    ):
        """Refer to SQLiteIndexedTar for all supported arguments and options."""

        options = kwargs.copy()

        self._open_file = None
        if isinstance(fo, str):
            # Implement URL chaining such as when calling fsspec.open("ratar://bar::file://single-file.tar").
            if target_protocol:
                self._open_file = fsspec.open(fo, protocol=target_protocol, **target_options)
                # Set the TAR file name so that the index can be found/stored accordingly.
                if target_protocol == 'file':
                    options['tarFileName'] = fo
                    if 'indexFilePath' not in options:
                        options['indexFilePath'] = fo + ".index.sqlite"
                    if 'writeIndex' not in options:
                        options['writeIndex'] = True
                if isinstance(self._open_file, fsspec.core.OpenFiles):
                    self._open_file = self._open_file[0]
                fo = self._open_file.open()
            else:
                options['tarFileName'] = fo
                if 'writeIndex' not in options:
                    options['writeIndex'] = True
                fo = None

        if fo:
            options['fileObject'] = fo

        super().__init__(SQLiteIndexedTar(**options))


# Only in case the entry point hooks in the pyproject.toml are not working for some reason.
fsspec.register_implementation("ratar", SQLiteIndexedTarFileSystem, clobber=True)



# pylint: disable=no-member,abstract-method
# Disable pylint errors. See https://github.com/fsspec/filesystem_spec/issues/1678

import http
import os
import re
import stat
import sys
import traceback
import warnings
from pathlib import Path
from typing import IO, Dict, Iterable, List, Union

from ratarmountcore.compressions import COMPRESSION_BACKENDS, check_for_split_file
from ratarmountcore.formats import FILE_FORMATS, FileFormatID
from ratarmountcore.StenciledFile import JoinedFileFromFactory
from ratarmountcore.utils import CompressionError, RatarmountError

from . import MountSource
from .archives import ARCHIVE_BACKENDS
from .compositing.singlefile import SingleFileMountSource
from .formats.folder import FolderMountSource
from .formats.fsspec import FSSpecMountSource
from .formats.git import GitMountSource

try:
    import fsspec
    import fsspec.implementations.http
    import fsspec.utils
except ImportError:
    fsspec = None  # type: ignore

try:
    from sshfs import SSHFileSystem

    class FixedSSHFileSystem(SSHFileSystem):
        protocols = ("sftp", "ssh", "scp")
        cachable = False

        def open(self, *args, **kwargs):
            # Note that asycnssh SSHFile does/did not implement seekable correctly!
            # https://github.com/fsspec/sshfs/pull/50
            result = super().open(*args, **kwargs)
            result.seekable = lambda: True  # type:ignore
            return result

except ImportError:
    FixedSSHFileSystem = None  # type: ignore

try:
    from webdav4.fsspec import WebdavFileSystem
except ImportError:
    WebdavFileSystem = None  # type: ignore

try:
    from dropboxdrivefs import DropboxDriveFileSystem

    class FixedDropboxDriveFileSystem(DropboxDriveFileSystem):
        def info(self, url, **kwargs):
            if url == ('/', ''):
                return {'size': 0, 'name': '/', 'type': 'directory'}
            return super().info(url, **kwargs)

except ImportError:
    FixedDropboxDriveFileSystem = None  # type: ignore


def _open_git_mount_source(url: str) -> Union[MountSource, IO[bytes], str]:
    splitURI = url.split('://', 1)
    if len(splitURI) <= 1 or splitURI[0] != 'git':
        raise RatarmountError("Expected URL starting with git://")

    if not GitMountSource.enabled:
        raise RatarmountError(
            "Detected git:// URL but GitMountSource could not be loaded. Please ensure that pygit2 is installed."
        )

    splitRepositoryPath = splitURI[1].split(':', 1)
    repositoryPath = splitRepositoryPath[0] if len(splitRepositoryPath) > 1 else None
    remainder = splitRepositoryPath[-1]

    splitReference = remainder.split('@', 1)
    reference = splitReference[0] if len(splitReference) > 1 else None
    pathInsideRepository = splitReference[-1]

    mountSource = GitMountSource(repositoryPath, reference=reference)
    if pathInsideRepository:
        fileInfo = mountSource.lookup(pathInsideRepository)
        if not fileInfo:
            raise RatarmountError(
                f"The path {pathInsideRepository} in the git repository specified via '{url}' does not exist!"
            )

        if stat.S_ISDIR(fileInfo.mode):
            mountSource.prefix = pathInsideRepository
        else:
            # In the future it might be necessary to extend the lifetime of mountSource by adding it as
            # a member of the opened file, but not right now.
            return mountSource.open(fileInfo)

    return mountSource


def _open_sshfs_mount_source(url: str) -> Union[MountSource, IO[bytes], str]:
    if FixedSSHFileSystem is None:
        raise RatarmountError("Cannot open with sshfs module because it seems to not be installed!")

    # Note that fsspec.implementations.ssh did not use ~/.ssh/config!
    # That's one of the many reasons why fsspec/sshfs based on asyncssh is used instead of paramiko.
    fs = FixedSSHFileSystem(**FixedSSHFileSystem._get_kwargs_from_urls(url))  # pytype: disable=attribute-error

    # Remove one leading / in order to add support for relative paths. E.g.:
    #   ssh://127.0.0.1/relative/path
    #   ssh://127.0.0.1//home/user/relative/path
    path = fsspec.utils.infer_storage_options(url)['path']
    if path.startswith("/"):
        path = path[1:]
    if not path:
        path = "."

    if not fs.exists(path):
        raise RatarmountError(f"Cannot open URL: {url} because the remote path: {path} does not exist!")
    # Note that the resulting file object has a .fs member for correct lifetime tracking.
    return fs.open(path) if fs.isfile(path) else FSSpecMountSource(fs, path)


def try_open_url(url, printDebug: int) -> Union[MountSource, IO[bytes], str]:
    splitURI = url.split('://', 1)
    protocol = splitURI[0] if len(splitURI) > 1 else ''
    if not protocol:
        raise RatarmountError(f"Expected to be called with URL containing :// but got: {url}")

    if printDebug >= 3:
        print(f"[Info] Try to open URL: {url}")

    if protocol == 'file':
        return splitURI[1]

    if protocol == 'git':
        return _open_git_mount_source(url)

    if FixedSSHFileSystem is not None and protocol in FixedSSHFileSystem.protocols:
        return _open_sshfs_mount_source(url)

    if not fsspec:
        raise RatarmountError(
            "An fsspec URL was detected but fsspec is not installed. Install it with: pip install ratarmount[fsspec]"
        )

    if printDebug >= 3:
        print("[Info] Try to open with fsspec")

    if protocol == 'ftp' and sys.version_info < (3, 9):
        url_to_fs = fsspec.url_to_fs if hasattr(fsspec, 'url_to_fs') else fsspec.core.url_to_fs
        # Suppress warning about (default!) encoding not being supported for Python<3.9 -.-.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fileSystem, path = url_to_fs(url)
    elif protocol == 'webdav':
        # WebDAV needs special handling because we need to decide between HTTP and HTTPS and because of:
        # https://github.com/skshetry/webdav4/issues/197
        if not WebdavFileSystem:
            raise RatarmountError(f"Install the webdav4 Python package to mount {protocol}://.")

        matchedURI = re.match("(?:([^:/]*):([^@/]*)@)?([^/]*)(.*)", splitURI[1])
        if not matchedURI:
            raise RatarmountError(
                "Failed to match WebDAV URI of the format webdav://[user:password@]host[:port][/path]\n"
                "If your user name or password contains special characters such as ':/@', then use the environment "
                "variables WEBDAV_USER and WEBDAV_PASSWORD to specify them."
            )
        username, password, baseURL, path = matchedURI.groups()
        if path is None:
            path = ""
        if username is None and 'WEBDAV_USER' in os.environ:
            username = os.environ.get('WEBDAV_USER')
        if password is None and 'WEBDAV_PASSWORD' in os.environ:
            password = os.environ.get('WEBDAV_PASSWORD')
        auth = None if username is None or password is None else (username, password)

        def check_for_https(url):
            try:
                connection = http.client.HTTPSConnection(url, timeout=2)
                connection.request("HEAD", "/")
                return bool(connection.getresponse())
            except Exception as exception:
                if printDebug >= 3:
                    print("[Info] Determined WebDAV URL to not use HTTP instead HTTPS because of:", exception)
                return False

        transportProtocol = "https" if check_for_https(baseURL) else "http"
        fileSystem = WebdavFileSystem(f"{transportProtocol}://{baseURL}", auth=auth)
    elif protocol == 'dropbox':
        # Dropbox needs special handling because there is no way to specify the token and because
        # there are some obnoxius intricacies regarding ls and stat of the root folder.
        if FixedDropboxDriveFileSystem is None:
            raise RatarmountError(f"Install the dropboxdrivefs Python package to mount {protocol}://.")

        dropboxToken = os.environ.get('DROPBOX_TOKEN', None)
        if not dropboxToken:
            raise RatarmountError(
                "Please set the DROPBOX_TOKEN environment variable to mount dropbox:// URLs. "
                "Please refer to the ratarmount online ReadMe or to the DropBox documentation for creating a token."
            )

        fileSystem = FixedDropboxDriveFileSystem(token=dropboxToken)
        path = splitURI[1]
        # Dropbox requires all paths to start with /, so simply add it
        # instead of making each user run into this problem.
        if path and not path.startswith('/'):
            path = '/' + path
        # Dropbox also does not like trailing / -.-. God is it super finicky.
        # dropbox.exceptions.ApiError: ApiError('12345', GetMetadataError('path', LookupError('malformed_path', None)))
        path = path.rstrip('/')
    else:
        url_to_fs = fsspec.url_to_fs if hasattr(fsspec, 'url_to_fs') else fsspec.core.url_to_fs
        fileSystem, path = url_to_fs(url)

    if printDebug >= 3:
        print("[Info] Opened filesystem:", fileSystem)

    # Note that http:// URLs are always files. Folders are only regex-parsed HTML files!
    # By checking with isdir instead of isfile, we give isdir a higher precedence.
    # Also note that isdir downloads the whole file!
    # https://github.com/fsspec/filesystem_spec/issues/1707
    if isinstance(fileSystem, fsspec.implementations.http.HTTPFileSystem):
        info = fileSystem.info(path)
        if info.get('mimetype', None) == 'text/html' and fileSystem.isdir(path):
            return FSSpecMountSource(fileSystem, path)
    elif fileSystem.isdir(path):
        return FSSpecMountSource(fileSystem, path)

    if not fileSystem.exists(path):
        raise RatarmountError(f"Opening URL {url} failed because path {path} does not exist on remote!")

    # This open call can fail with FileNotFoundError, IsADirectoryError, and probably others.
    result = fileSystem.open(path)  # pylint: disable=no-member

    # Avoid resource leaks, e.g., when the seek check fails.
    oldDel = getattr(result, '__del__', None)

    def new_del():
        if callable(oldDel):
            oldDel()
        result.close()

    result.__del__ = new_del

    # Check that seeking works. May fail when, e.g., the HTTP server does not support range requests.
    # Use https://github.com/danvk/RangeHTTPServer for testing purposes because
    # "python3 -m http.server 9000" does not have range support. Use "python3 -m RangeHTTPServer 9000".
    result.seek(1)
    result.read(1)
    result.seek(0)

    return result


def _matches_extension(fileName: str, formats: Iterable[FileFormatID]) -> bool:
    return any(
        fileName.lower().endswith('.' + extension.lower())
        for formatId in formats
        for extension in FILE_FORMATS[formatId].extensions
    )


def find_backends_by_extension(fileName: str) -> List[str]:
    return [backend for backend, info in ARCHIVE_BACKENDS.items() if _matches_extension(fileName, info.formats)] + [
        info.delegatedArchiveBackend
        for _, info in COMPRESSION_BACKENDS.items()
        if _matches_extension(fileName, info.formats)
    ]


def open_mount_source(fileOrPath: Union[str, IO[bytes], os.PathLike], **options) -> MountSource:
    printDebug = int(options.get("printDebug", 0)) if isinstance(options.get("printDebug", 0), int) else 0

    if isinstance(fileOrPath, str) and '://' in fileOrPath:
        openedURL = try_open_url(fileOrPath, printDebug=printDebug)

        # If the URL pointed to a folder, return a MountSource, else open the returned file object as an archive.
        if isinstance(openedURL, MountSource):
            return openedURL

        # Add tarFileName argument so that mounting a TAR file via SSH can create a properly named index
        # file inside ~/.cache/ratarmount.
        if not isinstance(openedURL, str) and 'tarFileName' not in options:
            options['tarFileName'] = fileOrPath

        fileOrPath = openedURL

    autoPrioritizedBackends: List[str] = []
    joinedFileName = ''
    if isinstance(fileOrPath, (str, os.PathLike)):
        path = Path(fileOrPath)
        if not path.exists():
            raise RatarmountError(f"Mount source does not exist: {fileOrPath!s}")

        if path.is_dir():
            return FolderMountSource('.' if str(fileOrPath) == '.' else path.resolve())

        splitFileResult = check_for_split_file(str(fileOrPath))
        if splitFileResult:
            filesToJoin = splitFileResult[0]
            joinedFileName = os.path.basename(filesToJoin[0]).rsplit('.', maxsplit=1)[0]
            if 'indexFilePath' not in options or not options['indexFilePath']:
                options['indexFilePath'] = filesToJoin[0] + ".index.sqlite"
            # https://docs.python.org/3/faq/programming.html
            # > Why do lambdas defined in a loop with different values all return the same result?
            fileOrPath = JoinedFileFromFactory(
                [(lambda file=file: open(file, 'rb')) for file in filesToJoin]  # type: ignore
            )
        else:
            fileOrPath = str(fileOrPath)
            autoPrioritizedBackends = find_backends_by_extension(fileOrPath)
    else:
        autoPrioritizedBackends = find_backends_by_extension(options.get('tarFileName', ''))

    prioritizedBackends = options.get("prioritizedBackends", [])
    triedBackends = set()
    # Map user-specified backend prioritization such as rapidgzip, indexed_bzip2, ... to tarfile,
    # which actually undoes those.
    mapToArchiveBackend: Dict[str, str] = {
        backend: info.delegatedArchiveBackend for backend, info in COMPRESSION_BACKENDS.items()
    }

    for name in prioritizedBackends + autoPrioritizedBackends + list(ARCHIVE_BACKENDS.keys()):
        name = mapToArchiveBackend.get(name, name)
        if name in triedBackends:
            continue
        triedBackends.add(name)
        if name not in ARCHIVE_BACKENDS:
            if printDebug >= 1:
                print(f"[Info] Skipping unknown archive backend: {name}")
            continue

        try:
            if printDebug >= 3:
                print(f"[Info] Try to open with {name}")
            result = ARCHIVE_BACKENDS[name].open(fileOrPath, **options)
            if result:
                if printDebug >= 2:
                    print(f"[Info] Opened archive with {name} backend.")
                return result
        except Exception as exception:
            if printDebug >= 2:
                print(f"[Info] Trying to open with {name} raised an exception:", exception)
            if printDebug >= 3:
                traceback.print_exc()

            try:
                if hasattr(fileOrPath, 'seek'):
                    fileOrPath.seek(0)  # type: ignore
            except Exception as seekException:
                if printDebug >= 1:
                    print("[Info] seek(0) raised an exception:", seekException)
                if printDebug >= 2:
                    traceback.print_exc()

    if joinedFileName and not isinstance(fileOrPath, (str, os.PathLike)):
        return SingleFileMountSource(joinedFileName, fileOrPath)

    raise CompressionError(f"Archive to open ({fileOrPath!s}) has unrecognized format!")



"""
Contains some quick format checks. Should not try to import large dependencies.
If the dependency is necessary, the check function should be moved into the respective MountSource implementation.
It basically is something like libmagic (Linux 'file' command), but only for the types we actually need.
There seem to be no nice, simple, widespread, and maintained Python package for this.

https://github.com/h2non/filetype.py                    MIT  ~700 stars    Last Release: 2022-11-02
    Small, dependency-free, fast Python package to infer binary file types checking the magic numbers signature
https://github.com/riad-azz/py-file-type                MIT  1 star        Last Release: 2023-04-22
     Wrapper for python-magic that includes necessary bin files.
https://github.com/enchant97/python-file-type-guesser   MIT  1 star        Last Release: 2021-12-20
    Guess a file's type from its content and extension.
    This is intended to be a pure Python implementation of something like python-magic.
https://github.com/dveselov/python-libmagic             MIT  ~30 stars     Last Release: 2016-07-07
    Python bindings to libmagic

The situation for file type detection Python packages is abysmal.
And I already skipped as many other low-effort packages on PyPI.

See:
 - https://en.wikipedia.org/wiki/List_of_file_signatures
 - https://en.wikipedia.org/wiki/List_of_archive_formats
"""

import dataclasses
import enum
import struct
import tarfile
import zipfile
from typing import IO, Callable, Dict, List, Optional, Set, Tuple, Union


class FileFormatID(enum.Enum):
    # fmt: off
    # Archive Formats (bundles more than one file)
    # "Normal" Archive formats with compression
    SEVEN_ZIP        = 0x101
    RAR              = 0x102
    ZIP              = 0x103
    SQLAR            = 0x104
    CAB              = 0x105
    AIX_SMALL        = 0x106
    AR               = 0x107
    XAR              = 0x108
    CPIO             = 0x109
    # "TAR"-like compression formats without compression
    TAR              = 0x201
    ASAR             = 0x202
    # Read-Only File systems
    SQUASHFS         = 0x301
    ISO9660          = 0x302
    # Fully-fledged file systems
    FAT              = 0x401
    EXT4             = 0x402
    # Other archive  formats
    RATARMOUNT_INDEX = 0x901
    WARC             = 0x902

    # Compression formats (compresses a single file / stream)
    BZIP2            = 0x1001
    GZIP             = 0x1002
    XZ               = 0x1003
    ZSTANDARD        = 0x1004
    ZLIB             = 0x1005
    LZ4              = 0x1006
    LZMA             = 0x1007

    GRZIP            = 0x1021
    LRZIP            = 0x1022
    LZIP             = 0x1023
    LZOP             = 0x1024
    RPM              = 0x1025
    UU               = 0x1026
    Z                = 0x1027
    # fmt: on


FID = FileFormatID


def is_tar(fileobj: IO[bytes], encoding: str = tarfile.ENCODING) -> bool:
    try:
        # r| will only open uncompressed TAR files and even allows unseekable ones!
        with tarfile.open(fileobj=fileobj, mode='r|', encoding=encoding):
            return True
    except (tarfile.ReadError, tarfile.CompressionError):
        pass
    return False


def find_asar_header(fileobj: IO[bytes]) -> Tuple[int, int, int]:
    """Return triple (start of header JSON, size of header JSON, start of file objects)"""
    # https://github.com/electron/asar/issues/128
    # https://github.com/electron/asar
    # > UInt32: header_size | String: header | Bytes: file1 | ...
    # > The header_size and header are serialized with Pickle class
    # > The header is a JSON string, and the header_size is the size of header's Pickle object.
    # Numbers are in little endian. Pickling writes the data, prepends a uint32 before and pads to 4 B offsets.
    # Because header_size is always 4 bytes, the implicit "magic" / redundant bytes of this file format are:
    # \x04\x00\x00\x00
    # Furthermore because pickling "header" already prepends the length of "header", the actual "header_size"
    # is completely redundant and can also be used as some kind of integrity check...
    # I don't understand where the fourth 32-bit number comes from. For some reason, the header seems to be
    # pickled twice, contrary to the ReadMe!?
    # https://github.com/electron/asar/issues/16
    # https://github.com/electron/asar/issues/226
    # https://knifecoat.com/Posts/ASAR+Format+Spec
    # -> Nice rant about the format as I have to concur about the usage of pickle-js.
    ASAR_MAGIC_SIZE = 4 * 4
    sizeOfPickledSize, sizeOfPickledPickledPickledHeader, sizeOfPickledPickledHeader, sizeOfPickledHeader = (
        struct.unpack('<LLLL', fileobj.read(ASAR_MAGIC_SIZE))
    )
    if sizeOfPickledSize != 4:
        raise ValueError("First magic bytes quadruplet does not match SQLAR!")
    if sizeOfPickledPickledPickledHeader != sizeOfPickledPickledHeader + 4:
        raise ValueError("Second magic bytes quadruplet does not match SQLAR!")
    padding = (4 - sizeOfPickledHeader % 4) % 4
    if sizeOfPickledPickledHeader != sizeOfPickledHeader + padding + 4:
        raise ValueError("Third magic bytes quadruplet does not match SQLAR!")

    dataOffset = ASAR_MAGIC_SIZE + sizeOfPickledHeader + padding

    # It would be nice if we could check that dataOffset < file size, but we cannot get that (cheaply)
    # in case of compressed files or for large headers and because "seek(offset)" will always return that
    # offset even if it is > file size, so we would actually have to read the data.

    return ASAR_MAGIC_SIZE, sizeOfPickledHeader, dataOffset


def is_asar(fileobj: IO[bytes]) -> bool:
    offset = fileobj.tell()
    try:
        # Reading the header and checking it to be correct JSON is to expensive for large archives.
        # The unnecessary pickling already should introduce enough redundancy for checks and detection.
        find_asar_header(fileobj)
        return True
    except Exception:
        pass
    finally:
        fileobj.seek(offset)
    return False


def is_squashfs(fileobj: IO[bytes]) -> bool:
    offset = fileobj.tell()
    try:
        # https://dr-emann.github.io/squashfs/squashfs.html#_the_superblock
        magicBytes = fileobj.read(4)
        if magicBytes != b"hsqs":
            return False

        _inodeCount, _modificationTime, blockSize, _fragmentCount = struct.unpack('<IIII', fileobj.read(4 * 4))
        compressor, blockSizeLog2, _flags, _idCount, major, minor = struct.unpack('<HHHHHH', fileobj.read(6 * 2))
        # root_inode, bytes_used, id_table, xattr_table, inode_table, dir_table, frag_table, export_table =
        # struct.unpack('<QQQQQQQQ', fileobj.read(8 * 8))

        # The size of a data block in bytes. Must be a power of two between 4096 (4k) and 1048576 (1 MiB).
        # log2 4096 = 12, log2 1024*1024 = 20
        if blockSizeLog2 < 12 or blockSizeLog2 > 20 or 2**blockSizeLog2 != blockSize:
            return False

        if major != 4 or minor != 0:
            return False

        # Compressions: 0:None, 1:GZIP, 2:LZMA, 3:LZO, 4:XZ, 5:LZ4, 6:ZSTD
        if compressor > 6:
            return False

    finally:
        fileobj.seek(offset)

    return True


def find_squashfs_offset(fileobj: IO[bytes], maxSkip=1024 * 1024) -> int:
    """
    Looks for the SquashFS superblock, which can be at something other than offset 0 for AppImage files.
    """
    # https://dr-emann.github.io/squashfs/squashfs.html#_the_superblock
    if is_squashfs(fileobj):
        return 0

    oldOffset = fileobj.tell()
    try:
        magic = b"hsqs"
        data = fileobj.read(maxSkip + len(magic))
        magicOffset = 0
        while True:
            magicOffset = data.find(magic, magicOffset + 1)
            if magicOffset < 0 or magicOffset >= len(data):
                break
            fileobj.seek(magicOffset)
            if is_squashfs(fileobj):
                return magicOffset
    finally:
        fileobj.seek(oldOffset)

    return -1


def _is_cpio(fileobj: IO[bytes]) -> bool:
    # http://justsolve.archiveteam.org/wiki/Cpio
    # https://github.com/libarchive/libarchive/blob/6110e9c82d8ba830c3440f36b990483ceaaea52c/libarchive/
    #   archive_read_support_format_cpio.c#L272
    firstBytes = fileobj.read(5)
    return firstBytes == b'07070' or firstBytes[:2] in [b'\x71\xc7', b'\xc7\x71']


def _is_iso9660(fileobj: IO[bytes]) -> bool:
    # https://www.iso.org/obp/ui/#iso:std:iso:9660:ed-1:v1:en
    # http://www.brankin.com/main/technotes/Notes_ISO9660.htm
    # https://en.wikipedia.org/wiki/ISO_9660
    # https://en.wikipedia.org/wiki/Optical_disc_image
    offset = 32 * 1024 + 1
    udfOffset = 38 * 1024 + 1
    buffer = fileobj.read(max(offset, udfOffset) + 4)
    return buffer[offset : offset + 5] == b'CD001' or buffer[udfOffset : udfOffset + 4] == b'NSR0'


def _check_zlib_header(fileobj: IO[bytes]) -> bool:
    header = fileobj.read(2)
    cmf = header[0]
    if cmf & 0xF != 8:
        return False
    if cmf >> 4 > 7:
        return False
    flags = header[1]
    if ((cmf << 8) + flags) % 31 != 0:
        return False
    usesDictionary = ((flags >> 5) & 1) != 0
    return not usesDictionary


def _is_bzip2(fileobj: IO[bytes]) -> bool:
    return fileobj.read(4)[:3] == b'BZh' and fileobj.read(6) == (0x314159265359).to_bytes(6, 'big')


def _check_lz4_header(fileobj: IO[bytes]) -> bool:
    SKIPPABLE_FRAME_MAGIC = 0x184D2A50
    (magic,) = struct.unpack('<L', fileobj.read(4))

    # https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md#skippable-frames
    while magic & 0xFFFF_FFF0 == SKIPPABLE_FRAME_MAGIC:
        (frame_size,) = struct.unpack('<L', fileobj.read(4))
        fileobj.seek(fileobj.tell() + frame_size)
        (magic,) = struct.unpack('<L', fileobj.read(4))

    # https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md#general-structure-of-lz4-frame-format
    return magic == 0x184D2204


def _check_zstandard_header(fileobj: IO[bytes]) -> bool:
    SKIPPABLE_FRAME_MAGIC = 0x184D2A50
    (magic,) = struct.unpack('<L', fileobj.read(4))

    # https://github.com/facebook/zstd/blob/dev/doc/zstd_compression_format.md#skippable-frames
    while magic & 0xFFFF_FFF0 == SKIPPABLE_FRAME_MAGIC:
        (frame_size,) = struct.unpack('<L', fileobj.read(4))
        fileobj.seek(fileobj.tell() + frame_size)
        (magic,) = struct.unpack('<L', fileobj.read(4))

    return magic == 0xFD2FB528


@dataclasses.dataclass
class FileFormatInfo:
    # Extensions without the initial '.'
    extensions: List[str]
    # If the first bytes of a format are constant, then they should be stated here.
    # This is used to build a lookup table for faster format detection as a first stage to thin down
    # the formats to actually call checkHeader for.
    # Some formats like ZIP (only has magic footer) and ISO9660 don't have a magic footer!
    magicBytes: Optional[bytes]
    # File format checkers should err on the side of false positives because else the file
    # will be rejected without actually trying to open it!
    # For simplicity, assume that all of these checkHeader will read from the current offset and be
    # at an arbitrary offset after the check. Normally, the offset should be larger than before in order
    # to mostly work with non-seekable files (although how is the caller supposed to parse the format after
    # an opaque check that skips an arbitrary amount of header?).
    checkHeader: Optional[Callable[[IO[bytes]], bool]] = None


ARCHIVE_FORMATS: Dict[FileFormatID, FileFormatInfo] = {
    # "Normal" Archive formats with compression
    FID.SEVEN_ZIP: FileFormatInfo(['7z', '7zip'], b'7z\xbc\xaf\x27\x1c'),
    FID.RAR: FileFormatInfo(['rar'], b'Rar!\x1a\x07'),
    # is_zipfile might yields some false positives, we want it to err on the positive side.
    # See: https://bugs.python.org/issue42096
    FID.ZIP: FileFormatInfo(['zip'], None, zipfile.is_zipfile),
    # SQLAR can be encrypted, for which it will not have any magic bytes! The first 16 B are the salt.
    # lambda x: x.read(16) == b'SQLite format 3\x00'
    FID.SQLAR: FileFormatInfo(['sqlar'], None, None),
    # https://download.microsoft.com/download/4/d/a/4da14f27-b4ef-4170-a6e6-5b1ef85b1baa/[ms-cab].pdf
    FID.CAB: FileFormatInfo(['cab'], b'MSCF'),
    # https://www.ibm.com/docs/en/aix/7.2.0?topic=formats-ar-file-format-small
    FID.AIX_SMALL: FileFormatInfo(['ar'], b'<aiaff>\n'),
    # https://www.ibm.com/docs/en/aix/7.2.0?topic=formats-ar-file-format-big#ar_big
    # https://en.wikipedia.org/wiki/Ar_(Unix)
    FID.AR: FileFormatInfo(['a', 'ar', 'lib'], b'!<arch>\n'),
    FID.XAR: FileFormatInfo(['xar'], b'xar!'),
    FID.CPIO: FileFormatInfo(['cpio'], None, _is_cpio),
    # "TAR"-like compression formats without compression
    FID.TAR: FileFormatInfo(['tar'], None, is_tar),
    FID.ASAR: FileFormatInfo(['asar'], b'\x04\x00\x00\x00', is_asar),
    # Read-Only File systems
    FID.SQUASHFS: FileFormatInfo(['squashfs', 'AppImage', 'snap'], None, lambda x: find_squashfs_offset(x) >= 0),
    FID.ISO9660: FileFormatInfo(['iso'], None, _is_iso9660),
    # Fully-fledged file systems
    FID.FAT: FileFormatInfo(['fat', 'img', 'dd', 'fat12', 'fat16', 'fat32', 'raw'], None),
    FID.EXT4: FileFormatInfo(['ext4', 'img', 'dd', 'raw'], None),
    # Other archive formats
    FID.RATARMOUNT_INDEX: FileFormatInfo(['index.sqlite'], b'SQLite format 3\x00'),
    # https://www.iso.org/standard/68004.html
    # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#file-and-record-model
    FID.WARC: FileFormatInfo(['warc'], b'WARC/1.'),
}

COMPRESSION_FORMATS: Dict[FileFormatID, FileFormatInfo] = {
    FID.BZIP2: FileFormatInfo(['bz2', 'bzip2'], b'BZh', _is_bzip2),
    FID.GZIP: FileFormatInfo(['gz', 'gzip'], b'\x1f\x8b'),
    FID.XZ: FileFormatInfo(['xz'], b"\xfd7zXZ\x00"),
    FID.ZSTANDARD: FileFormatInfo(['zst', 'zstd', 'pzstd'], None, _check_zstandard_header),
    FID.ZLIB: FileFormatInfo(['zz', 'zlib'], None, _check_zlib_header),
    # https://github.com/libarchive/libarchive/blob/6110e9c82d8ba830c3440f36b990483ceaaea52c/libarchive/
    # archive_read_support_filter_grzip.c#L46
    # It is almost impossible to find the original sources or a specification:
    # https://t2sde.org/packages/grzip
    # ftp://ftp.ac-grenoble.fr/ge/compression/grzip-0.3.0.tar.bz2
    FID.GRZIP: FileFormatInfo(['grz', 'grzip'], b'GRZipII\x00\x02\x04:)'),
    # https://github.com/ckolivas/lrzip/blob/master/doc/magic.header.txt
    FID.LRZIP: FileFormatInfo(['lrz', 'lrzip'], b'LRZI\x00'),
    # https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md#general-structure-of-lz4-frame-format
    FID.LZ4: FileFormatInfo(['lz4'], None, _check_lz4_header),
    # https://www.ietf.org/archive/id/draft-diaz-lzip-09.txt
    FID.LZIP: FileFormatInfo(['lz', 'lzip'], b'LZIP\x01'),
    # https://github.com/jljusten/LZMA-SDK/blob/master/DOC/lzma-specification.txt
    # https://github.com/frizb/FirmwareReverseEngineering/blob/master/IdentifyingCompressionAlgorithms.md#lzma
    FID.LZMA: FileFormatInfo(['lzma'], b'\x5d\x00\x00'),
    FID.LZOP: FileFormatInfo(['lzo', 'lzop'], b'\x89\x4c\x5a\x4f\x00\x0d\x0a\x1a\x0a'),
    # https://refspecs.linuxbase.org/LSB_4.1.0/LSB-Core-generic/LSB-Core-generic/pkgformat.html
    FID.RPM: FileFormatInfo(['rpm'], b'\xed\xab\xee\xdb'),
    # https://en.wikipedia.org/wiki/Uuencoding
    FID.UU: FileFormatInfo(['uu'], b'begin '),
    # https://github.com/file/file/blob/master/magic/Magdir/compress
    FID.Z: FileFormatInfo(['Z'], b'\x1f\x9d'),
}


FILE_FORMATS = {**ARCHIVE_FORMATS, **COMPRESSION_FORMATS}

# Check that all defined format IDs have FileFormatInfo, so that we can assume FORMATS[FID] to not fail.
for _formatInfo in FileFormatID:
    assert _formatInfo in FILE_FORMATS, f"Missing file format information for: {_formatInfo}"


# Maps the first 2 bytes to possible candidates. 2 B was chosen because most formats have
# at least 2 B of constant magic and if the table was implemented in contiguous memory, it would be nicely
# addressable by a 16-bit index (the 2 B). Statistically, this cuts down the number of expensive checkHeader
# calls by 65536.
_MAGIC_BYTES_TO_FORMATS: Dict[bytes, List[FileFormatID]] = {}
_FORMATS_WITHOUT_MAGIC_BYTES: List[FileFormatID] = []


def recompute_cached_magic_bytes():
    """Should be called when injecting new file formats from outsfide."""
    for fid, info in FILE_FORMATS.items():
        if info.magicBytes is None or len(info.magicBytes) < 2:
            _FORMATS_WITHOUT_MAGIC_BYTES.append(fid)
        else:
            firstTwoBytes = info.magicBytes[:2]
            if firstTwoBytes not in _MAGIC_BYTES_TO_FORMATS:
                _MAGIC_BYTES_TO_FORMATS[firstTwoBytes] = []
            _MAGIC_BYTES_TO_FORMATS[firstTwoBytes].append(fid)


recompute_cached_magic_bytes()


def might_be_format(fileobj: IO[bytes], fid: Union[FileFormatID, FileFormatInfo]) -> bool:
    formatInfo = fid if isinstance(fid, FileFormatInfo) else FILE_FORMATS[fid]
    oldOffset = fileobj.tell()
    try:
        if formatInfo.magicBytes and fileobj.read(len(formatInfo.magicBytes)) != formatInfo.magicBytes:
            return False
        if formatInfo.checkHeader:
            fileobj.seek(oldOffset)
            return formatInfo.checkHeader(fileobj)
    finally:
        fileobj.seek(oldOffset)

    # No magic bytes and no header check, reject it because there seems to be the necessary module to read
    # that format missing anyway.
    return bool(formatInfo.magicBytes or formatInfo.checkHeader)


def detect_formats(fileobj: IO[bytes]) -> Set[FileFormatID]:
    oldOffset = fileobj.tell()
    try:
        firstTwoBytes = fileobj.read(2)
        # I don't think there is any file format that can be recognized if the file is smaller than 2 B.
        if len(firstTwoBytes) < 2:
            return set()

        formatsToTest = _MAGIC_BYTES_TO_FORMATS.get(firstTwoBytes, [])
    finally:
        fileobj.seek(oldOffset)

    return {fid for fid in formatsToTest + _FORMATS_WITHOUT_MAGIC_BYTES if might_be_format(fileobj, fid)}


def replace_format_check(fid: FileFormatID, checkHeader: Optional[Callable[[IO[bytes]], bool]] = None):
    for formats in [ARCHIVE_FORMATS, COMPRESSION_FORMATS, FILE_FORMATS]:
        if fid in formats:
            info = dataclasses.asdict(formats[fid])
            info['checkHeader'] = checkHeader
            formats[fid] = FileFormatInfo(**info)



import contextlib
import io
import os
import stat
import threading
import time
from typing import IO, Any, Dict, Iterable, Optional, Union, cast

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.StenciledFile import RawStenciledFile, StenciledFile
from ratarmountcore.utils import overrides


class SingleFileMountSource(MountSource):
    """MountSource exposing a single file as a mount source."""

    # It makes no sense to merge this class into SubvolumesMountSource because of the file locking and file size
    # for each file that needs to be saved and it would require adding if-else into almost every method in
    # SubvolumesMountSource. Having a simple SingleFileMountSource to be used with SubvolumesMountSource is far
    # more elegant. The only difference might be how to handle multiple mount points/files per parent folder.
    # This is currently not possible because SingleFileMountSource implies a single folder containing a single file.
    # It could be emulated with UnionMountSource or by extending SingleFileMountSource to support multiple files,
    # but then it would add inconsistencies when these multiple files have different full paths / implied parents...

    def __init__(self, path: str, fileobj: IO[bytes]):
        """
        fileobj: The given file object to be mounted. It may be advisable for this file object to be unbuffered
                 because opening file objects via this mount source will add additional buffering if not disabled.
        """
        self.path = os.path.normpath('/' + path).lstrip('/')
        if self.path.endswith('/') or not self.path:
            raise ValueError("File object must belong to a non-folder path!")

        self.fileObjectLock = threading.Lock()
        self.fileobj = fileobj
        self.mtime = int(time.time())
        self.size: int = self.fileobj.seek(0, io.SEEK_END)

        fileno = None
        with contextlib.suppress(Exception):
            fileno = self.fileobj.fileno()

        self._statfs = {}
        if fileno is not None:
            statfs = os.fstat(fileno)
            self._statfs = {
                'f_bsize': statfs.st_blksize,
                'f_frsize': statfs.st_blksize,
                'f_blocks': statfs.st_blocks,
                'f_bfree': 0,
                'f_bavail': 0,
                'f_ffree': 0,
                'f_favail': 0,
            }

    def _create_file_info(self):
        # This must be a function and cannot be cached into a member in order to avoid userdata being a shared list!
        # fmt: off
        return FileInfo(
            size     = self.size,
            mtime    = self.mtime,
            mode     = int(0o777 | stat.S_IFREG),
            linkname = '',
            uid      = 0,
            gid      = 0,
            userdata = [],
        )
        # fmt: on

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        pathWithSlash = (path.strip('/') + '/').lstrip('/')  # append / to be able to use startswith correctly
        if self.path.startswith(pathWithSlash):
            return [self.path[len(pathWithSlash) :].split('/', maxsplit=1)[0]]
        return None

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        pathWithSlash = (path.strip('/') + '/').lstrip('/')  # append / to be able to use startswith correctly
        if self.path.startswith(pathWithSlash):
            # fmt: off
            return FileInfo(
                size     = 0,
                mtime    = self.mtime,
                mode     = int(0o777 | stat.S_IFDIR),
                linkname = '',
                uid      = 0,
                gid      = 0,
                userdata = [],
            )
            # fmt: on

        return self._create_file_info() if path.strip('/') == self.path else None

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        if fileInfo != self._create_file_info():
            raise ValueError("Only files may be opened!")

        # Use StenciledFile so that the returned file objects can be independently seeked!
        if buffering == 0:
            return cast(
                IO[bytes],
                RawStenciledFile(fileStencils=[(self.fileobj, 0, self.size)], fileObjectLock=self.fileObjectLock),
            )
        return cast(
            IO[bytes],
            StenciledFile(
                fileStencils=[(self.fileobj, 0, self.size)],
                fileObjectLock=self.fileObjectLock,
                bufferSize=io.DEFAULT_BUFFER_SIZE if buffering <= 0 else buffering,
            ),
        )

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        return self._statfs.copy()

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.fileobj.close()

    def join_threads(self):
        if hasattr(self.fileobj, 'join_threads'):
            self.fileobj.join_threads()



import base64
import collections
import contextlib
import importlib
import io
import math
import os
import pathlib
import platform
import sys
import types
from pathlib import Path
from typing import Dict, Generic, Iterable, List, Optional, TypeVar, Union, get_type_hints

try:
    import importlib.metadata
except ImportError:
    import importlib_metadata

    imeta: types.ModuleType = importlib_metadata
else:
    imeta = importlib.metadata


class RatarmountError(Exception):
    """Base exception for ratarmount module."""


class IndexNotOpenError(RatarmountError):
    """Exception for operations executed on a closed index database."""


class InvalidIndexError(RatarmountError):
    """Exception for indexes being invalid, outdated, or created with different arguments."""


class MismatchingIndexError(RatarmountError):
    """Exception for indexes being created by a different backend."""


class CompressionError(RatarmountError):
    """Exception for trying to open files with unsupported compression or unavailable decompression module."""


def overrides(parentClass):
    """Simple decorator that checks that a method with the same name exists in the parent class"""
    # I tried typing.override (Python 3.12+), but support for it does not seem to be ideal (yet)
    # and portability also is an issue. https://github.com/google/pytype/issues/1915 Maybe in 3 years.

    def overrider(method):
        if platform.python_implementation() == 'PyPy':
            return method

        assert method.__name__ in dir(parentClass)
        parentMethod = getattr(parentClass, method.__name__)
        assert callable(parentMethod)

        if os.getenv('RATARMOUNT_CHECK_OVERRIDES', '').lower() not in ('1', 'yes', 'on', 'enable', 'enabled'):
            return method

        # Example return of get_type_hints:
        # {'path': <class 'str'>,
        #  'return': typing.Union[typing.Iterable[str],
        #                         typing.Dict[str, ratarmountcore.MountSource.FileInfo], NoneType]}
        parentTypes = get_type_hints(parentMethod)
        # If the parent is not typed, e.g., fusepy, then do not show errors for the typed derived class.
        for argument, argumentType in get_type_hints(method).items():
            if argument in parentTypes:
                parentType = parentTypes[argument]
                assert argumentType == parentType, f"{method.__name__}: {argument}: {argumentType} != {parentType}"

        return method

    return overrider


class _DummyContext:
    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


def ceil_div(dividend, divisor):
    return -(dividend // -divisor)


KeyType = TypeVar('KeyType')
ValueType = TypeVar('ValueType')


class LRUCache(Generic[KeyType, ValueType]):
    """
    Provides a cache behaving like dictionary with a limited size, which automatically deletes
    least-recently-used keys-value pairs when the size limit has been reached.
    """

    # Does not inherit from dict to ensure that there aren't any methods not overwritten and causing inconsistencies.
    # E.g. copy would return a dicft instead of LRUCache and so on.
    def __init__(self, size: int = 10):
        self.size = size
        self.lastUsed: List[KeyType] = []
        self.data: Dict[KeyType, ValueType] = {}

    def _refresh(self, key: KeyType):
        if key in self.lastUsed:
            self.lastUsed.remove(key)
        self.lastUsed.append(key)

    def __getitem__(self, key: KeyType):
        self._refresh(key)
        return self.data.__getitem__(key)

    def get(self, key: KeyType, default: Optional[ValueType] = None):
        if key in self.lastUsed:
            self.lastUsed.remove(key)
            self.lastUsed.append(key)
        return self.data.get(key, default)

    def __setitem__(self, key: KeyType, value: ValueType):
        self.data.__setitem__(key, value)

        self._refresh(key)
        while self.data.__len__() > self.size:
            self.data.__delitem__(self.lastUsed.pop(0))

    def __delitem__(self, key: KeyType):
        self.data.__delitem__(key)
        if key in self.lastUsed:
            self.lastUsed.remove(key)

    def __contains__(self, key: KeyType):
        return self.data.__contains__(key)

    def __len__(self):
        return self.data.__len__()

    def __repr__(self):
        return self.data.__repr__()

    def __str__(self):
        return self.data.__str__()

    def clear(self):
        self.data.clear()
        self.lastUsed.clear()

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def setdefault(self, key: KeyType, default: ValueType):
        self._refresh(key)
        return self.data.setdefault(key, default)

    def __bool__(self):
        return bool(self.data)


class Prefetcher:
    def __init__(self, memorySize):
        self.lastFetched = []
        self.memorySize = memorySize

    def fetch(self, value):
        if value in self.lastFetched:
            self.lastFetched.remove(value)
        self.lastFetched.append(value)
        while len(self.lastFetched) > self.memorySize:
            self.lastFetched.pop(0)

    def prefetch(self, maximumToPrefetch) -> Iterable:
        if not self.lastFetched or maximumToPrefetch <= 0:
            return []

        consecutiveCount = 0
        values = self.lastFetched[::-1]
        for i, j in zip(values[0:-1], values[1:]):
            if i == j + 1:
                consecutiveCount += 1
            else:
                break

        # I want an exponential progression like: logStep**consecutiveCount with the boundary conditions:
        # logStep**0 = 1 (mathematically true for any logStep because consecutiveCount was chosen to fit)
        # logStep**maxConsecutiveCount = maximumToPrefetch
        #   => logStep = exp(ln(maximumToPrefetch)/maxConsecutiveCount)
        #   => logStep**consecutiveCount = exp(ln(maximumToPrefetch) * consecutiveCount/maxConsecutiveCount)
        prefetchCount = round(math.exp(math.log(maximumToPrefetch) * consecutiveCount / (self.memorySize - 1)))
        return range(self.lastFetched[-1] + 1, self.lastFetched[-1] + 1 + prefetchCount)


ALPHA = ''.join(chr(ord('a') + i) for i in range(ord('z') - ord('a') + 1))
DIGITS = ''.join(chr(ord('0') + i) for i in range(ord('9') - ord('0') + 1))
HEX = DIGITS + ALPHA[:6]


def is_latin_alpha(text: str):
    return text and all(ord('a') <= ord(c) <= ord('z') for c in text)


def is_latin_digit(text: str):
    return text and all(ord('0') <= ord(c) <= ord('9') for c in text)


def is_latin_hex_alpha(text: str):
    return text and all(ord('0') <= ord(c) <= ord('9') or ord('a') <= ord(c) <= ord('f') for c in text)


def format_number(i: int, base: str, length: int = 0):
    if len(base) <= 1:
        raise ValueError("Base alphabet must contain more than one letter!")

    result = ''
    while i > 0 or length > 0 or not result:
        result += base[i % len(base)]
        i = i // len(base)
        length = length - 1
    return result[::-1]


def distribution_contains_file(distribution, path: str) -> bool:
    if not distribution.files:
        return False

    for file in distribution.files:
        if not path.endswith(str(file)):
            continue

        try:
            pathlib.Path(path).relative_to(file.locate())
            return True
        except ValueError:
            return False
    return False


def get_module(module: Union[str, types.ModuleType]) -> Optional[types.ModuleType]:
    if isinstance(module, types.ModuleType):
        return module

    if module not in sys.modules:
        with contextlib.suppress(ImportError):
            importlib.import_module(module)
    return sys.modules.get(module, None)


def find_module_version(moduleOrName: Union[str, types.ModuleType]) -> Optional[str]:
    module = get_module(moduleOrName)
    if not module:
        return None

    # zipfile has no __version__ attribute and PEP 396 ensuring that was rejected 2021-04-14
    # in favor of 'version' from importlib.metadata which does not even work with zipfile.
    # Probably, because zipfile is a built-in module whose version would be the Python version.
    # https://www.python.org/dev/peps/pep-0396/
    # The "python-xz" project is imported as an "xz" module, which complicates things because
    # there is no generic way to get the "python-xz" name from the "xz" runtime module object
    # and importlib.metadata.version will require "python-xz" as argument.
    # Note that even when querying the version with importlib.metadata.version, it can return
    # a different version than the actually imported module if some import tricks were done
    # like manipulating sys.path to import a non-installed module.
    # All in all, this really gets on my nerves and I wished that PEP 396 would have been accepted.
    # Currently, it feels like work has been shifted from the maintainer side to the user side.
    # See below the kinds of handstands we have to do to even just get the unreliable package
    # name from the module name in order to query the version. It's mental.
    # And note that importlib.metadata has only been introduced in Python 3.8 and has been
    # provisional until including Python 3.9, meaning it can still take years to be available
    # and stable on all systems or I have to add yet another dependency just to get the damn version.
    version = getattr(module, '__version__', None)
    if version:
        return str(version)

    moduleFilePath = getattr(module, '__file__', None)
    if moduleFilePath:
        for distribution in imeta.distributions():
            try:
                if distribution_contains_file(distribution, moduleFilePath) and 'Version' in distribution.metadata:
                    return distribution.metadata['Version']
            except Exception:
                pass

    return None


def is_on_slow_drive(filePath: str):
    # TODO determine whether the whole file or most of it has been cached:
    #      https://serverfault.com/questions/278454/is-it-possible-to-list-the-files-that-are-cached
    #      https://github.com/mxmlnkn/rapidgzip/issues/13#issuecomment-1592856413
    # TODO make it work on Windows: https://devblogs.microsoft.com/oldnewthing/20201023-00/?p=104395
    try:
        device = os.stat(filePath).st_dev
        if Path(f"/sys/dev/block/{os.major(device)}:{os.minor(device)}/queue/rotational").read_bytes().strip() == b"1":
            return True
    except Exception:
        pass
    return False


class FixedRawIOBase(io.RawIOBase):
    @overrides(io.RawIOBase)
    def readall(self) -> bytes:
        # It is necessary to implement this, or else the io.RawIOBase.readall implementation would use
        # io.DEFAULT_BUFFER_SIZE (8 KiB). Notably, this would ignore the block size configured in BufferedReader,
        # when calling read(-1) on it because it thinks that raw.readall is a fast-path, but in this case is ~100x
        # slower than 4 MiB reads equal to the Lustre-advertised block size.
        # https://github.com/python/cpython/issues/85624
        chunks = []
        while True:
            result = self.read()
            if not result:
                break
            chunks.append(result)
        return b"".join(chunks)


def get_xdg_cache_home():
    # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    # > $XDG_CACHE_HOME defines the base directory relative to which user-specific non-essential data files should
    # > be stored. If $XDG_CACHE_HOME is either not set or empty, a default equal to $HOME/.cache should be used.
    path = os.environ.get('XDG_CACHE_HOME', '')
    home = os.path.expanduser("~/")
    if path.startswith(home):
        path = "~/" + path[len(home) :]
    return path or os.path.join('~', '.cache')


def decode_unpadded_base64(data: str) -> bytes:
    return base64.b64decode(data + '=' * ((4 - len(data) % 4) % 4))


def remove_duplicates_stable(iterable: Iterable):
    seen = set()
    deduplicated = []
    for x in iterable:
        if x not in seen:
            deduplicated.append(x)
            seen.add(x)
    return deduplicated


def determine_recursion_depth(recursive: bool = False, recursionDepth: Optional[int] = None, **_) -> int:
    """
    'recursionDepth' has higher precedence than 'recursive' if specified.
    Can be called directly i.e. f(a,b) or with kwargs f(**options) for which unknown keys will be ignored.
    Returns a real recursion depth. -1 or any negative has no special meaning, but it should effectively
    result in no recursion.
    """
    return (sys.maxsize if recursive else 0) if recursionDepth is None else recursionDepth


def compute_entropy(data: bytes) -> float:
    if not data:
        return 0.0

    # https://en.wikipedia.org/wiki/Cross-entropy
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.entropy.html
    # > The cross entropy can be calculated as the sum of the entropy and relative entropy.
    # The entropy we compute is basically the number of bits required to encode the message optimally,
    # e.g., with Huffman Coding or, better, with asymmetric numeral systems.
    # I.e., for random data we except the result to be close to 8, e.g. cut off > 7.5, which would correspond
    # to a compression ratio of 0.9375.
    # Note that we would need at least 256 values, or else 256 - N values will be 0 and will therefore
    # reduce entropy! More generically, t should be a multiple of 256 to get more accurate entropy estimations!
    # Tests with SQLAR encrypted data:
    #   N = 10   -> 3.32
    #   N = 100  -> 6.38
    #   N = 384  -> 7.49
    #   N = 1000 -> 7.81
    #
    #   N = 256  -> 7.24
    #   N = 512  -> 7.61
    #   N = 1024 -> 7.81
    #   N = 2048 -> 7.916
    #   N = 4096 -> 7.950
    #   N = 8192 -> 7.977
    # Note that the information entropy is completely ordering-independent!
    # I.e., a sequence like \x00\x01\x02...\xFF will have entropy like random data even though it clearly is not!
    # The first idea coming to mind to encode position, would to take the difference between neighboring bytes.
    # This would still result in the same value range but we could test for sequences, and all linearly related
    # sequences such as \x00\x02\x04..\xFC\xFE.
    # It would even find repeated words because any repeated word would also repeat the same differences between
    # consecutive letters.
    # I really have trouble constructing data that this method wouldn't catch. I guess quadratic sequences like
    # \x01\x02\x04\x08\... could work with overflow but at that point we would go into the direction of pseudorandom
    # generators like (quadratic) linear congruential generators.
    probabilities = [byteFrequency / len(data) for byteFrequency in collections.Counter(data).values()]
    return -sum(p * math.log2(p) for p in probabilities)


def is_random(data: bytes) -> bool:
    if len(data) <= 1:
        return False

    # See benchmarks/scripts/testEntropy to heuristically find a lower bound for entropy of random-distributed data.
    # The longer data is the closer it will get to 8 bits entropy. The difference goes to 0 proportional to 1/length.
    # Larger m make randomness detection more lenient. m=16 follows approximately the median when repeating with
    # different random data, so m should at least be >16.
    # We could do any number of tests for randomness: https://en.wikipedia.org/wiki/Statistical_randomness
    # Some of them might even fail for encrypted data if the encryption is assumedly not strong enough or adds magic
    # bytes at periodic locations.
    # Does not work very well for small number of bytes! Should at least give 1024. A multiple of 256 is preferable.
    m = 40

    def convergence_rate(n: int):
        return 8 * (1 + m) / (n + m)

    def is_in_threshold(dataToTest: bytes):
        return 8 - compute_entropy(dataToTest) < convergence_rate(len(dataToTest))

    diffData = bytes((data[i + 1] - data[i] + 256) % 256 for i in range(len(data) - 1))
    return is_in_threshold(data) and is_in_threshold(diffData)



"""Ratarmount Core

This is the backend of ratarmount. It is intended to be used as a library.

This library offers an interface which is sufficient to work with FUSE for
read-only access.

The most common usecase should be covered by the open_mount_source factory
function. For more information, see the ratarmountcore.mountsource submodule.

Example:

    from ratarmountcore.mountsources.factory import open_mount_source

    archive = open_mount_source("foo.tar", recursive=True)
    archive.list("/")
    info = archive.lookup("/bar")

    print("Contents of /bar:")
    with archive.open(info) as file:
        print(file.read())
"""

from .version import __version__



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import hashlib
import os
import stat
import sys

import pytest

try:
    import sqlcipher3
except ImportError:
    sqlcipher3 = None  # type:ignore

from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.sqlar import SQLARMountSource  # noqa: E402


class TestSQLARMountSource:
    @staticmethod
    @pytest.mark.parametrize('path', ['nested-tar.sqlar', 'nested-tar-compressed.sqlar', 'encrypted-nested-tar.sqlar'])
    @pytest.mark.parametrize('passwords', [None, [b"foo"], [b"'; DROP TABLE sqlar;", b"foo", b"c"]])
    def test_password(path, passwords):
        if 'encrypted' in path and (not sqlcipher3 or not passwords):
            return
        with copy_test_file(path) as tmpPath, SQLARMountSource(tmpPath, passwords=passwords) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo', '/foo/lighter.tar']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)

            with mountSource.open(mountSource.lookup('/foo/fighter/ufo')) as file:
                assert file.read() == b'iriya\n'

            with mountSource.open(mountSource.lookup('/foo/lighter.tar')) as file:
                if 'encrypted' in path:
                    assert hashlib.md5(file.read()).hexdigest() == "2a06cc391128d74e685a6cb7cfe9f94d"
                else:
                    assert hashlib.md5(file.read()).hexdigest() == "4dfaddf7e55e48097d34e03936223a50"

    # TODO Does not use SQLiteIndex backend because it is already a SQLite database and it would seem redundant.
    #      Therefore 'transform' does not work.



import contextlib
import datetime
import json
import re
import stat
import sys
import tarfile
import zipfile
from timeit import default_timer as timer
from typing import IO, Any, Dict, List, Optional, Tuple, Union

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.utils import overrides

try:
    # Importing this patches the zipfile module as a "side" effect!
    import fast_zip_decryption  # pylint: disable=unused-import  # noqa: F401
except (ImportError, Exception):
    with contextlib.suppress(ImportError, Exception):
        import fastzipfile  # pylint: disable=unused-import  # noqa: F401


class ZipMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes]],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[List[str]]       = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        printDebug             : int                       = 0,
        indexMinimumFileCount  : int                       = 1000,
        transform              : Optional[Tuple[str, str]] = None,
        **options
    ) -> None:
        # fmt: on
        if 'zipfile' not in sys.modules:
            raise RuntimeError("Did not find the zipfile module. Please use Python 3.7+.")

        # fmt: off
        self.fileObject             = zipfile.ZipFile(fileOrPath, 'r')
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.printDebug             = printDebug
        self.options                = options
        self.transformPattern       = transform
        # fmt: on

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        ZipMountSource._find_password(self.fileObject, options.get("passwords", []))
        self.files = {info.header_offset: info for info in self.fileObject.infolist()}

        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                printDebug=self.printDebug,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='ZipMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        isFileObject = not isinstance(fileOrPath, str)

        if self.index.index_is_loaded():
            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding', 'transformPattern']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)

    def _convert_to_row(self, info: "zipfile.ZipInfo") -> Tuple:
        mtime = datetime.datetime(*info.date_time, tzinfo=datetime.timezone.utc).timestamp() if info.date_time else 0

        # According to section 4.5.7 in the .ZIP file format specification, links are supported:
        # https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT
        # The Python zipfile module has no API for links: https://bugs.python.org/issue45286
        # However, the file mode exposes whether it's a link and the file mode is shown by ZipInfo.__repr__.
        # For that, it uses the OS-dependent external_attr member. See also the ZIP specification on that:
        # > 4.4.15 external file attributes: (4 bytes)
        # >   The mapping of the external attributes is host-system dependent (see 'version made by').
        # >   For MS-DOS, the low order byte is the MS-DOS directory attribute byte.
        # >   If input came from standard input, this field is set to zero.

        # file_redir is (type, flags, target) or None. Only tested for type == RAR5_XREDIR_UNIX_SYMLINK.
        linkname = ""
        mode = (info.external_attr >> 16) & 0o777
        if stat.S_ISLNK(info.external_attr >> 16):
            linkname = self.fileObject.read(info).decode()
            mode = mode | stat.S_IFLNK
        else:
            mode = mode | (stat.S_IFDIR if info.is_dir() else stat.S_IFREG)

        path, name = SQLiteIndex.normpath(self.transform(info.filename)).rsplit("/", 1)

        # Currently, this is unused. The index only is used for getting metadata. (The data offset
        # is already determined and written out in order to possibly speed up reading of encrypted
        # files by implementing the decryption ourselves.)
        # The data offset is deprecated again! Collecting it can add a huge overhead for large zip files
        # because we have to seek to every position and read a few bytes from it. Furthermore, it is useless
        # by itself anyway. We don't even store yet how the data is compressed or encrypted, so we would
        # have to read the local header again anyway!
        dataOffset = 0

        # fmt: off
        fileInfo : Tuple = (
            path              ,  # 0  : path
            name              ,  # 1  : file name
            info.header_offset,  # 2  : header offset
            dataOffset        ,  # 3  : data offset
            info.file_size    ,  # 4  : file size
            mtime             ,  # 5  : modification time
            mode              ,  # 6  : file mode / permissions
            0                 ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            linkname          ,  # 8  : linkname
            0                 ,  # 9  : user ID
            0                 ,  # 10 : group ID
            False             ,  # 11 : is TAR (unused?)
            False             ,  # 12 : is sparse
            False             ,  # 13 : is generated (parent folder)
            0                 ,  # 14 : recursion depth
        )
        # fmt: on

        return fileInfo

    def _create_index(self) -> None:
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()
        self.index.set_file_infos([self._convert_to_row(info) for info in self.fileObject.infolist()])

        # Resort by (path,name). This one-time resort is faster than resorting on each INSERT (cache spill)
        if self.printDebug >= 2:
            print("Resorting files by path ...")

        self.index.finalize()

        t1 = timer()
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} took {t1 - t0:.2f}s")

    @staticmethod
    def _find_password(fileobj: "zipfile.ZipFile", passwords):
        # If headers are encrypted, then infolist will simply return an empty list!
        files = fileobj.infolist()
        if not files:
            for password in passwords:
                fileobj.setpassword(password)
                files = fileobj.infolist()
                if files:
                    return password

        # If headers are not encrypted, then try out passwords by trying to open the first file.
        files = [file for file in files if not file.is_dir() and file.file_size > 0]
        if not files:
            return None

        for password in [None, *passwords]:
            fileobj.setpassword(password)
            try:
                with fileobj.open(files[0]) as file:
                    file.read(1)
                return password
            except Exception:
                pass

        raise RuntimeError("Could not find a matching password!")

    @overrides(SQLiteIndexMountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        super().__exit__(exception_type, exception_value, exception_traceback)
        self.fileObject.close()

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        # I do not see any obvious option to zipfile.ZipFile to apply the specified buffer size.
        assert fileInfo.userdata
        extendedFileInfo = fileInfo.userdata[-1]
        assert isinstance(extendedFileInfo, SQLiteIndexedTarUserData)
        info = self.files[extendedFileInfo.offsetheader]
        assert isinstance(info, zipfile.ZipInfo)
        # CPython's zipfile module does handle multiple file objects being opened and reading from the
        # same underlying file object concurrently by using a _SharedFile class that even includes a lock.
        # Very nice!
        # https://github.com/python/cpython/blob/a87c46eab3c306b1c5b8a072b7b30ac2c50651c0/Lib/zipfile/__init__.py#L1569
        return self.fileObject.open(info, 'r')  # https://github.com/pauldmccarthy/indexed_gzip/issues/85

    def _check_metadata(self, metadata: Dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(
                json.loads(metadata['arguments']), self, argumentsToCheck=['encoding', 'transformPattern']
            )

        if 'backendName' not in metadata:
            self.index.try_to_open_first_file(lambda path: self.open(self.lookup(path)))



import contextlib
import json
import os
import re
import stat
import tarfile
import threading
from timeit import default_timer as timer
from typing import IO, Any, Dict, List, Optional, Tuple, Union, cast

from ratarmountcore.formats import find_asar_header
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.StenciledFile import RawStenciledFile, StenciledFile
from ratarmountcore.utils import overrides


# https://www.electronjs.org/docs/latest/glossary#asar
# > ASAR stands for Atom Shell Archive Format. An asar archive is a simple tar-like format
# > that concatenates files into a single file.
#
# Example header:
# {
#    "files": {
#       "tmp": {
#          "files": {}
#       },
#       "usr" : {
#          "files": {
#            "bin": {
#              "files": {
#                "ls": {
#                  "offset": "0",
#                  "size": 100,
#                  "executable": true,
#                  "integrity": {
#                    "algorithm": "SHA256",
#                    "hash": "...",
#                    "blockSize": 1024,
#                    "blocks": ["...", "..."]
#                  }
#                }
#              }
#            }
#          }
#       },
#       "etc": {
#          "files": {
#            "hosts": {
#              "offset": "200",
#              "size": 32,
#              "integrity": {
#                 "algorithm": "SHA256",
#                 "hash": "...",
#                 "blockSize": 1024,
#                 "blocks": ["...", "..."]
#               }
#            }
#          }
#       }
#    }
# }
#
# There are no permissions, symbolic links (but hard links are possible), or file times.
class ASARMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes]],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[List[str]]       = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        printDebug             : int                       = 0,
        indexMinimumFileCount  : int                       = 1000,
        transform              : Optional[Tuple[str, str]] = None,
        **options
    ) -> None:
        self.isFileObject           = not isinstance(fileOrPath, str)
        self.fileObject             = open(fileOrPath, 'rb') if isinstance(fileOrPath, str) else fileOrPath
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.printDebug             = printDebug
        self.options                = options
        self.transformPattern       = transform
        # fmt: on

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        # Try to open file
        self._headerOffset, self._headerSize, self._dataOffset = find_asar_header(self.fileObject)
        self.fileObject.seek(self._headerOffset + self._headerSize - 1)
        if not self.fileObject.read(1):
            raise ValueError(
                f"Failed to read the full ASAR header sized {self._headerSize} from offset {self._headerOffset}!"
            )

        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                printDebug=self.printDebug,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='ASARMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        # Try to get block size from the real opened file.
        self.blockSize = 512
        with contextlib.suppress(Exception):
            self.blockSize = os.fstat(self.fileObject.fileno()).st_blksize

        self.fileObjectLock = threading.Lock()

        # Load or create index (copy-paste)

        if self.index.index_is_loaded():
            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not self.isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding', 'transformPattern']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)

    def _convert_to_row(self, fullPath, entry: Dict[str, Any], dataOffset: int) -> Tuple:
        path, name = SQLiteIndex.normpath(self.transform(fullPath)).rsplit("/", 1)

        isFile = 'offset' in entry and 'size' in entry
        isDirectory = 'files' in entry
        assert isFile != isDirectory
        mode = 0o777 | (stat.S_IFDIR if isDirectory else stat.S_IFREG)

        offset = dataOffset + int(entry.get("offset", "0"))
        size = int(entry.get("size", "0"))

        # fmt: off
        fileInfo : Tuple = (
            path              ,  # 0  : path
            name              ,  # 1  : file name
            0                 ,  # 2  : header offset
            offset            ,  # 3  : data offset
            size              ,  # 4  : file size
            0                 ,  # 5  : modification time
            mode              ,  # 6  : file mode / permissions
            0                 ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            ""                ,  # 8  : linkname
            0                 ,  # 9  : user ID
            0                 ,  # 10 : group ID
            False             ,  # 11 : is TAR (unused?)
            False             ,  # 12 : is sparse
            False             ,  # 13 : is generated (parent folder)
            0                 ,  # 14 : recursion depth
        )
        # fmt: on

        return fileInfo

    def _create_index(self) -> None:
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()

        # Using StenciledFile instead of josn.loads(self.fileObject.read()) to avoid yet another copy in memory,
        # to roughly halve memory usage for very large JSONs. Note that the JSON size is limited to 4 GiB because
        # the size is 32-bit.
        header = json.load(StenciledFile([(self.fileObject, self._headerOffset, self._headerSize)]))

        # This code is complex to avoid recursion.
        toProcess = [("/", header)]
        fileInfos = [self._convert_to_row("/", header, self._dataOffset)]
        while toProcess:
            prefix, entry = toProcess.pop()
            children = [(prefix + "/" + path, nestedEntry) for path, nestedEntry in entry['files'].items()]
            # toProcess must only contain folder entries, which have a 'files' key mapping to a dictionary.
            fileInfos += [self._convert_to_row(path, nestedEntry, self._dataOffset) for path, nestedEntry in children]
            # Appending leads to depth-first traversal, while prepending leads to breadth-first traversal.
            # Depth-first should be slightly less memory intensive. Note that the dictionary should only store
            # a pointer to the shared mutable Python object, but the path strings will use up memory.
            toProcess += [(path, nestedEntry) for path, nestedEntry in children if 'files' in nestedEntry]

            if len(fileInfos) > 1000:
                self.index.set_file_infos(fileInfos)
                fileInfos = []

        if fileInfos:
            self.index.set_file_infos(fileInfos)

        # Resort by (path,name). This one-time resort is faster than resorting on each INSERT (cache spill)
        if self.printDebug >= 2:
            print("Resorting files by path ...")

        self.index.finalize()

        t1 = timer()
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} took {t1 - t0:.2f}s")

    @overrides(SQLiteIndexMountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        super().__exit__(exception_type, exception_value, exception_traceback)
        if not self.isFileObject:
            self.fileObject.close()

    def _open_stencil(self, offset: int, size: int, buffering: int) -> IO[bytes]:
        if buffering == 0:
            return cast(IO[bytes], RawStenciledFile([(self.fileObject, offset, size)], self.fileObjectLock))
        return cast(
            IO[bytes],
            StenciledFile(
                [(self.fileObject, offset, size)],
                self.fileObjectLock,
                bufferSize=self.blockSize if buffering == -1 else buffering,
            ),
        )

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        assert fileInfo.userdata
        extendedFileInfo = fileInfo.userdata[-1]
        assert isinstance(extendedFileInfo, SQLiteIndexedTarUserData)
        return self._open_stencil(extendedFileInfo.offset, fileInfo.size, buffering)

    def _check_metadata(self, metadata: Dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(
                json.loads(metadata['arguments']), self, argumentsToCheck=['encoding', 'transformPattern']
            )

        if 'backendName' not in metadata:
            self.index.try_to_open_first_file(lambda path: self.open(self.lookup(path)))



from typing import IO, Any, Dict, Iterable, Optional, Union

from ratarmountcore.formats import FileFormatID, replace_format_check
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.utils import overrides

try:
    import ext4
except ImportError:
    ext4 = None  # type: ignore


def is_ext4_image(fileObject) -> bool:
    if ext4 is None:
        return False

    offset = fileObject.tell()
    try:
        ext4.Volume(fileObject)
        return True
    except Exception:
        pass
    finally:
        fileObject.seek(offset)
    return False


replace_format_check(FileFormatID.EXT4, is_ext4_image)


class EXT4MountSource(MountSource):
    def __init__(self, fileOrPath: Union[str, IO[bytes]], encoding: str = 'utf8', **options) -> None:
        if ext4 is None:
            raise ImportError("Failed to find python4-ext4. Try: pip install ext4")

        self.encoding = encoding
        self.fileObject = open(fileOrPath, 'rb') if isinstance(fileOrPath, str) else fileOrPath
        self.fileObjectWasOpened = isinstance(fileOrPath, str)
        self.fileSystem = ext4.Volume(self.fileObject)
        self.options = options

    @staticmethod
    def _convert_ext4_directory_entry_to_file_info(inode) -> FileInfo:
        # fmt: off
        return FileInfo(
            size     = inode.i_size,
            mtime    = inode.i_mtime,
            mode     = inode.i_mode.value,
            linkname = "",  # TODO I don't see any data for links... in the Inode struct
            uid      = inode.i_uid,
            gid      = inode.i_gid,
            userdata = [inode.i_no],
        )
        # fmt: on

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def exists(self, path: str) -> bool:
        try:
            self.fileSystem.inode_at(path)
        except FileNotFoundError:
            return False
        return True

    def _list(self, path: str, getValue) -> Optional[Dict[str, Any]]:
        try:
            inode = self.fileSystem.inode_at(path)
        except FileNotFoundError:
            return None

        if not isinstance(inode, ext4.inode.Directory):
            return None

        return {
            entry.name.decode(self.encoding): getValue(self.fileSystem.inodes[entry.inode])
            for entry, _ in inode.opendir()
            if entry.name not in (b'.', b'..')
        }

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        return self._list(path, EXT4MountSource._convert_ext4_directory_entry_to_file_info)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        return self._list(path, lambda inode: inode.i_mode.value)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        try:
            return self._convert_ext4_directory_entry_to_file_info(self.fileSystem.inode_at(path))
        except FileNotFoundError:
            pass
        return None

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return 1

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        inode = fileInfo.userdata[-1]
        assert isinstance(inode, int)
        return self.fileSystem.inodes[inode].open()

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self.fileObjectWasOpened:
            self.fileObject.close()



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.SQLiteIndex import SQLiteIndex  # noqa: E402


class TestSQLiteIndexedTarParallelized:
    @staticmethod
    def test_normpath():
        normpath = SQLiteIndex.normpath
        assert normpath("/") == "/"
        assert normpath("//") == "/"
        assert normpath("///") == "/"

        assert normpath("a") == "/a"
        assert normpath("/a") == "/a"
        assert normpath("a/") == "/a"
        assert normpath("/a/") == "/a"
        assert normpath("//a//") == "/a"

        assert normpath(".") == "/"
        assert normpath("/.") == "/"
        assert normpath("./") == "/"
        assert normpath("/./") == "/"
        assert normpath("//.//") == "/"

        assert normpath("./././a/.././") == "/"
        assert normpath("../") == "/"
        assert normpath("../.././..") == "/"

    @staticmethod
    def test_query_normpath():
        normpath = SQLiteIndex._query_normpath
        assert normpath("/") == "/"
        assert normpath("//") == "/"
        assert normpath("///") == "/"

        assert normpath("a") == "/a"
        assert normpath("/a") == "/a"
        assert normpath("a/") == "/a"
        assert normpath("/a/") == "/a"
        assert normpath("//a//") == "/a"

        assert normpath(".") == "/"
        assert normpath("/.") == "/"
        assert normpath("./") == "/"
        assert normpath("/./") == "/"
        assert normpath("//.//") == "/"

        assert normpath("./././a/.././") == "/"
        assert normpath("../") == "/.."
        assert normpath("../.././..") == "/../../.."



import enum
import os
import stat
from typing import IO, Any, Dict, Iterable, List, Optional, Tuple, Union

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.utils import overrides


class FileType(enum.Enum):
    FILE = 0
    VERSIONS_FOLDER = 1


class FileVersionLayer(MountSource):
    """
    This bind mount like layer makes it possible to access older file versions if there multiple ones in the given
    mount source. The interface provides for each file <file path> a hidden folder <file path.versions> containing
    all available versions.

    This class also resolves hardlinks. This functionality is mixed in here because self-referencing hardlinks
    should be resolved by showing older versions of a file and only this layer knows about file versioning.

    TODO If there already exists a file <file path.versions> then this special folder will not be available!
    """

    def __init__(self, mountSource: MountSource):
        self.mountSource: MountSource = mountSource

    def _decode_versions_path_api(self, filePath: str) -> Optional[Tuple[str, bool, int]]:
        """
        Do a loop over the parent path parts to resolve possible versions in parent folders.
        Note that multiple versions of a folder always are union mounted. So, for the path to a file
        inside those folders the exact version of a parent folder can simply be removed for lookup.
        Therefore, translate something like: /foo.version/3/bar.version/2/mimi.version/1 into
        /foo/bar/mimi.version/1
        This is possibly time-costly but requesting a different version from the most recent should
        be a rare occurrence and FUSE also checks all parent parts before accessing a file so it
        might only slow down access by roughly factor 2.
        """

        # TODO make it work for files ending with '.versions'.
        # Currently, this feature would be hidden by those files. But, I think this should be quite rare.
        # I could allow arbitrary amounts of dots like '....versions' but then it wouldn't be discernible
        # for ...versions whether the versions of ..versions or .versions file was requested. I could add
        # a rule for the decision, like ...versions shows the versions of .versions and ....versions for
        # ..versions, however, all of this might require an awful lot of file existence checking.
        # My first idea was to use hidden subfolders for each file like path/to/file/.versions/1 but FUSE
        # checks the parents in a path that they are directories first, so getattr or readdir is not even
        # called for path/to/file/.versions if path/to/file is not a directory.
        # Another alternative might be one hidden folder at the root for a parallel file tree, like
        # /.versions/path/to/file/3 but that runs into similar problems when trying to specify the file
        # version or if a .versions root directory exists.

        filePathParts = filePath.lstrip('/').split('/')
        filePath = ''
        pathIsSpecialVersionsFolder = False
        fileVersion = None  # Not valid if None or parentIsVersions is True
        for part in filePathParts:
            # Skip over the exact version specified
            if pathIsSpecialVersionsFolder:
                try:
                    fileVersion = int(part)
                    assert str(fileVersion) == part
                except Exception:
                    return None
                pathIsSpecialVersionsFolder = False
                continue

            # Simply append normal existing folders
            tmpFilePath = filePath + '/' + part
            if self.mountSource.lookup(tmpFilePath):
                filePath = tmpFilePath
                fileVersion = 0
                continue

            # If current path does not exist, check if it is a special versions path
            if part.endswith('.versions') and len(part) > len('.versions'):
                pathIsSpecialVersionsFolder = True
                fileVersion = 0
                filePath = tmpFilePath[: -len('.versions')]
                continue

            # Parent path does not exist and is not a versions path, so any subpaths also won't exist either
            return None

        if fileVersion is None:
            return None

        return filePath, pathIsSpecialVersionsFolder, (0 if pathIsSpecialVersionsFolder else fileVersion)

    @staticmethod
    def _is_hard_link(fileInfo: FileInfo) -> bool:
        # Note that S_ISLNK checks for symbolic links. Hardlinks (at least from tarfile)
        # return false for S_ISLNK but still have a linkname!
        return bool(not stat.S_ISREG(fileInfo.mode) and not stat.S_ISLNK(fileInfo.mode) and fileInfo.linkname)

    @staticmethod
    def _resolve_hard_links(mountSource: MountSource, path: str) -> Optional[FileInfo]:
        """path : Simple path. Should contain no special versioning folders!"""

        fileInfo = mountSource.lookup(path)
        if not fileInfo:
            return None

        resolvedPath = '/' + fileInfo.linkname.lstrip('/') if FileVersionLayer._is_hard_link(fileInfo) else None
        fileVersion = 0
        hardLinkCount = 0

        while resolvedPath and hardLinkCount < 128:  # For comparison, the maximum symbolic link chain in Linux is 40.
            # Link targets are relative to the mount source. That's why we need the mount point to get the full path
            # in respect to this mount source. And we must a file info object for this mount source, so we have to
            # get that using the full path instead of calling lookup on the deepest mount source.
            mountPoint, _, _ = mountSource.get_mount_source(fileInfo)

            resolvedPath = os.path.join(mountPoint, resolvedPath.lstrip('/'))

            if resolvedPath != path:  # noqa: SIM108
                # The file version is only of importance to resolve self-references.
                # It seems undecidable to me whether to return the given fileVersion or 0 here.
                # Returning 0 would feel more correct because the we switched to another file and the version
                # for that file is the most recent one.
                # However, resetting the file version to 0 means that if there is a cycle, i.e., two hardlinks
                # of different names referencing each other, than the file version will always be reset to 0
                # and we have no break condition, entering an infinite loop.
                # The most correct version would be to track the version of each path in a map and count up the
                # version per path.
                # TODO Is such a hardlink cycle even possible?!
                fileVersion = 0
            else:
                # If file is referencing itself, try to access earlier version of it.
                # The check for fileVersion against the total number of available file versions is omitted because
                # that check is done implicitly inside the mount sources lookup method!
                fileVersion = fileVersion + 1 if fileVersion >= 0 else fileVersion - 1

            path = resolvedPath
            fileInfo = mountSource.lookup(path, fileVersion)
            if not fileInfo:
                return None

            resolvedPath = '/' + fileInfo.linkname.lstrip('/') if FileVersionLayer._is_hard_link(fileInfo) else None
            hardLinkCount += 1

        return fileInfo

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return self.mountSource.is_immutable()

    def _list_wrapper(self, list_function, path: str):
        files = list_function(path)
        if files is not None:
            if isinstance(files, dict):
                for fileInfo in files.values():
                    if isinstance(fileInfo, FileInfo):
                        fileInfo.userdata.append(FileType.FILE)
            return files

        # If no folder was found, check whether the special .versions folder was requested
        try:
            result = self._decode_versions_path_api(path)
        except Exception:
            return None

        if not result:
            return None
        path, pathIsSpecialVersionsFolder, _ = result

        if not pathIsSpecialVersionsFolder:
            files = list_function(path)
            if isinstance(files, dict):
                for fileInfo in files.values():
                    if isinstance(fileInfo, FileInfo):
                        fileInfo.userdata.append(FileType.FILE)
            return files

        # Print all available versions of the file at filePath as the contents of the special '.versions' folder
        return [str(version + 1) for version in range(self.mountSource.versions(path))]

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        return self._list_wrapper(self.mountSource.list, path)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        return self._list_wrapper(self.mountSource.list_mode, path)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        """Resolves special file version specifications in the path."""

        assert fileVersion == 0

        fileInfo = FileVersionLayer._resolve_hard_links(self.mountSource, path)
        if fileInfo:
            fileInfo.userdata.append(FileType.FILE)
            return fileInfo

        # If no file was found, check if a special .versions folder to an existing file/folder was queried.
        versionsInfo = self._decode_versions_path_api(path)
        if not versionsInfo:
            return None
        path, pathIsSpecialVersionsFolder, fileVersion = versionsInfo

        # 2.) Check if the request was for the special .versions folder and return its contents or stats
        # At this point, path is assured to actually exist!
        if pathIsSpecialVersionsFolder:
            parentFileInfo = self.mountSource.lookup(path)
            assert parentFileInfo

            # fmt: off
            return FileInfo(
                size     = 0,
                mtime    = parentFileInfo.mtime,
                mode     = 0o777 | stat.S_IFDIR,
                linkname = "",
                uid      = parentFileInfo.uid,
                gid      = parentFileInfo.gid,
                userdata = [FileType.VERSIONS_FOLDER],
            )
            # fmt: on

        # 3.) At this point the request is for an actually older version of a file or folder
        fileInfo = self.mountSource.lookup(path, fileVersion=fileVersion)
        if fileInfo:
            fileInfo.userdata.append(FileType.FILE)
        return fileInfo

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return self.mountSource.versions(path)

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        fileType = fileInfo.userdata.pop()
        try:
            if fileType == FileType.FILE:
                return self.mountSource.open(fileInfo, buffering=buffering)
            raise FileNotFoundError(f"[FileVersionLayer.open] file info: {fileInfo}")
        finally:
            fileInfo.userdata.append(fileType)

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        fileType = fileInfo.userdata.pop()
        try:
            if fileType == FileType.FILE:
                return self.mountSource.read(fileInfo, size, offset)
            raise FileNotFoundError(f"[FileVersionLayer.read] file info: {fileInfo}")
        finally:
            fileInfo.userdata.append(fileType)

    @overrides(MountSource)
    def get_mount_source(self, fileInfo: FileInfo) -> Tuple[str, MountSource, FileInfo]:
        fileType = fileInfo.userdata.pop()
        try:
            if fileType == FileType.VERSIONS_FOLDER:
                return '/', self, fileInfo
            return self.mountSource.get_mount_source(fileInfo)
        finally:
            fileInfo.userdata.append(fileType)

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        fileType = fileInfo.userdata.pop()
        try:
            if fileType == FileType.VERSIONS_FOLDER:
                return []
            return self.mountSource.list_xattr(fileInfo)
        finally:
            fileInfo.userdata.append(fileType)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        fileType = fileInfo.userdata.pop()
        try:
            if fileType == FileType.VERSIONS_FOLDER:
                return None
            return self.mountSource.get_xattr(fileInfo, key)
        finally:
            fileInfo.userdata.append(fileType)

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.mountSource.__exit__(exception_type, exception_value, exception_traceback)

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        return self.mountSource.statfs()

    def join_threads(self):
        if hasattr(self.mountSource, 'join_threads'):
            self.mountSource.join_threads()



import concurrent.futures
import dataclasses
import itertools
import os
import platform
import struct
import sys
import traceback
from typing import IO, Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, cast

from .BlockParallelReaders import ParallelXZReader
from .formats import ARCHIVE_FORMATS, COMPRESSION_FORMATS, FID, FileFormatID, might_be_format
from .utils import (
    ALPHA,
    DIGITS,
    HEX,
    CompressionError,
    format_number,
    is_latin_alpha,
    is_latin_digit,
    is_latin_hex_alpha,
    is_on_slow_drive,
)

try:
    import indexed_gzip
except ImportError:
    indexed_gzip = None  # type: ignore

try:
    import indexed_zstd
except ImportError:
    indexed_zstd = None  # type: ignore
try:
    import lzmaffi

except ImportError:
    lzmaffi = None  # type: ignore

try:
    import xz
except ImportError:
    if 'xz' not in sys.modules:
        # For some reason, only this import triggers mypy. All the others are fine.
        # Should be something like Optional[Module] but there is no Module type.
        xz = None  # type: ignore

try:
    import rapidgzip
except ImportError:
    rapidgzip = None  # type: ignore

try:
    import zstandard
except ImportError:
    zstandard = None  # type: ignore


try:
    # Must be imported because find_available_backend checks for it to be in sys.modules!
    # Although, I'm unsure whether it gets implicitly added to sys.modules below when importing file_reader.
    # OSError can happen when dependencies are missing, e.g., libicuuc.so.74.
    import libarchive  # pylint: disable=unused-import
except Exception:
    libarchive = None  # type: ignore

try:
    from libarchive import file_reader as libarchive_file_reader
except Exception:

    def libarchive_file_reader(path):
        raise ImportError("Please install python-libarchive-c with: pip install libarchive-c")


TAR_CONTRACTED_EXTENSIONS: Dict[FileFormatID, List[str]] = {
    FID.BZIP2: ['tb2', 'tbz', 'tbz2', 'tz2'],
    FID.GZIP: ['taz', 'tgz'],
    FID.XZ: ['txz'],
    FID.ZSTANDARD: ['tzst'],
}


@dataclasses.dataclass
class CompressionBackendInfo:
    # Opens a file object from a path or file object and additional options (kwargs).
    open: Callable[..., IO[bytes]]
    # Supported file formats. These are for quick checks and prioritization based on file extension.
    formats: Set[FileFormatID]
    # If a format is suspected e.g. by extension or by a non-module dependent format check,
    # the modules listed here are checked and the module package name can be suggested to be installed.
    # Tuple: (module name, package name)
    requiredModules: List[Tuple[str, str]]
    delegatedArchiveBackend: str


COMPRESSION_BACKENDS: Dict[str, CompressionBackendInfo] = {
    'rapidgzip-bzip2': CompressionBackendInfo(
        (lambda x, parallelization=0: rapidgzip.IndexedBzip2File(x, parallelization=parallelization)),
        {FID.BZIP2},
        [('rapidgzip', 'rapidgzip')],
        'tarfile',
    ),
    'rapidgzip': CompressionBackendInfo(
        (lambda x, parallelization=1: rapidgzip.RapidgzipFile(x, parallelization=parallelization)),
        {FID.GZIP, FID.ZLIB},
        [('rapidgzip', 'rapidgzip')],
        'tarfile',
    ),
    'indexed_gzip': CompressionBackendInfo(
        (lambda x, parallelization=1: indexed_gzip.IndexedGzipFile(fileobj=x)),
        {FID.GZIP},
        [('indexed_gzip', 'indexed_gzip')],
        'tarfile',
    ),
    # Prioritize xz over lzmaffi
    'xz': CompressionBackendInfo(
        (lambda x, parallelization=1: cast(IO[bytes], xz.open(x))), {FID.XZ}, [('xz', 'python-xz')], 'tarfile'
    ),
    'lzmaffi': CompressionBackendInfo(
        (lambda x, parallelization=1: lzmaffi.open(x)), {FID.XZ}, [('lzmaffi', 'lzmaffi')], 'tarfile'
    ),
    'indexed_zstd': CompressionBackendInfo(
        (lambda x, parallelization=1: indexed_zstd.IndexedZstdFile(x.fileno())),
        {FID.ZSTANDARD},
        [('indexed_zstd', 'indexed_zstd')],
        'tarfile',
    ),
    'libarchive': CompressionBackendInfo(
        # TODO Does not work because libarchive_file_reader is some kind of context manager.
        #      This only affects detect_compression for undoing the compression in SQLiteIndexedTar,
        #      so not that important for now.
        (lambda x, parallelization=1: libarchive_file_reader(x)),
        {
            # Also supported by other backends, which should be used first.
            FID.BZIP2,
            FID.GZIP,
            FID.XZ,
            FID.ZSTANDARD,
            # Compression formats
            FID.GRZIP,
            FID.LRZIP,
            FID.LZ4,
            FID.LZIP,
            FID.LZMA,
            FID.LZOP,
            FID.RPM,
            FID.UU,
            FID.Z,
        },
        [('libarchive', 'libarchive-c')],
        'libarchive',
    ),
}


def find_available_backend(
    compression: FileFormatID,
    enabledBackends: Optional[List[str]] = None,
    prioritizedBackends: Optional[List[str]] = None,
) -> Optional[CompressionBackendInfo]:
    if prioritizedBackends is None:
        prioritizedBackends = []

    matchingBackends = [
        backend
        for backend, info in COMPRESSION_BACKENDS.items()
        if (enabledBackends is None or backend in enabledBackends) and compression in info.formats
    ]

    for backendName in prioritizedBackends:
        if backendName not in matchingBackends or backendName not in COMPRESSION_BACKENDS:
            continue
        backend = COMPRESSION_BACKENDS[backendName]
        if all(module in sys.modules for module, _ in backend.requiredModules):
            return backend

    for backendName in matchingBackends:
        if backendName in prioritizedBackends or backendName not in COMPRESSION_BACKENDS:
            continue
        backend = COMPRESSION_BACKENDS[backendName]
        if all(module in sys.modules for module, _ in backend.requiredModules):
            return backend

    return None


def strip_suffix_from_compressed_file(path: str) -> str:
    """Strips compression suffixes like .bz2, .gz, ..."""
    for formatInfo in COMPRESSION_FORMATS.values():
        for extension in formatInfo.extensions:
            if path.lower().endswith('.' + extension.lower()):
                return path[: -(len(extension) + 1)]
    return path


def strip_suffix_from_archive(path: str) -> str:
    """Strips extensions like .tar.gz or .gz or .tgz, .rar, .zip ..."""
    extensions = itertools.chain(
        (e for extensions in TAR_CONTRACTED_EXTENSIONS.values() for e in extensions),
        ('t' + e for formatInfo in COMPRESSION_FORMATS.values() for e in formatInfo.extensions),
        ('tar.' + e for formatInfo in COMPRESSION_FORMATS.values() for e in formatInfo.extensions),
        (e for formatInfo in COMPRESSION_FORMATS.values() for e in formatInfo.extensions),
        (e for formatInfo in ARCHIVE_FORMATS.values() for e in formatInfo.extensions),
    )
    for extension in extensions:
        if path.lower().endswith('.' + extension.lower()):
            return path[: -(len(extension) + 1)]
    return path


def has_matching_alphabets(a: str, b: str):
    return (
        (is_latin_alpha(a) and is_latin_alpha(b))
        or (is_latin_digit(a) and is_latin_digit(b))
        or (is_latin_hex_alpha(a) and is_latin_hex_alpha(b))
    )


def check_for_sequence(extensions: Iterable[str], numberFormatter: Callable[[int], str]) -> List[str]:
    suffixSequence: List[str] = []
    i = 0
    suffixLength = len(numberFormatter(0))

    while True:
        suffix = numberFormatter(i)
        if suffix in extensions:
            suffixSequence.append(suffix)
        elif i > 0 or len(suffix) != suffixLength:
            # We allow the extensions to start with 0 or 1.
            # So, even if the zeroth does not exist, do not break, instead also test 1.
            break
        i += 1

    return suffixSequence


def check_for_split_file(path: str) -> Optional[Tuple[List[str], str]]:
    """
    Returns the paths to all files belonging to the split and a string identifying the format.
    The latter is one of: '', 'x', 'a' to specify the numbering system: decimal, hexadecimal, alphabetical.
    The width and starting number can be determined from the extension of the first file path,
    which is returned in the first member of the tuple.
    """

    # Check for split files. Note that GNU coreutils' split 8.32 by default creates files in the form:
    #   xaa, xab, xac, ..., xba, ...
    # However, for now, don't support those. At least a dot should be used to reduce misidentification!
    # Split supports not only alphabetical, but also decimal, and hexadecimal numbering schemes and the
    # latter, both, start at 0, which is also quite unusual.
    # Most files I encountered in the wild used decimal suffixes starting with name.001.
    # Split can be customized a lot with these options:
    #   split [OPTION]... [FILE [PREFIX]]
    #   --suffix-length
    #   --additional-suffix=SUFFIX
    #   --numeric-suffixes[=FROM]
    #   --hex-suffixes[=FROM]
    #   --suffix-length
    # -> It seems like there is no way to specify alphabetical counting to begin at something other than a, aa, ...
    # To avoid false positives, these restrictions to split's options are made:
    #   - FROM is 0 or 1. Anything else is too aberrant to be supported.
    #   - PREFIX ends with a dot (.).

    # These character tests are necessary because Python's built-in isalpha, isdigit and such
    # all also return true for a lot of Unicode alternatives like for the Thai zero.

    folder, filename = os.path.split(os.path.realpath(path))  # Get file extensions
    if '.' not in filename:
        return None
    basename, extension = filename.rsplit('.', maxsplit=1)

    # Collect all other files in the folder that might belong to the same split.
    extensions = [name[len(basename) + 1 :] for name in os.listdir(folder) if name.startswith(basename + '.')]
    extensions = [e for e in extensions if has_matching_alphabets(e, extension)]
    if not extensions:
        return None
    assert extension in extensions

    # Note that even if something consists only of letters or only digits it still might be hexadecimal encoding!
    maxFormatSpecifier = ''
    maxExtensions: List[str] = []
    for formatSpecifier, baseDigits in [('a', ALPHA), ('0', DIGITS), ('x', HEX)]:
        extensionSequence = check_for_sequence(
            extensions, lambda i, baseDigits=baseDigits: format_number(i, baseDigits, len(extension))  # type: ignore
        )
        if len(extensionSequence) > len(maxExtensions):
            maxFormatSpecifier = formatSpecifier
            maxExtensions = extensionSequence

    if maxFormatSpecifier and len(maxExtensions) > 1:
        paths = [path.rsplit('.', maxsplit=1)[0] + '.' + extension for extension in maxExtensions]
        return paths, maxFormatSpecifier

    return None


def _compress_zstd(data):
    return zstandard.ZstdCompressor().compress(data)


def compress_zstd(filePath: str, outputFilePath: str, frameSize: int, parallelization: Optional[int] = None):
    """
    Compresses filePath into outputFilePath with one zstandard frame for each frameSize chunk of uncompressed data.
    """
    if not parallelization:
        parallelization = os.cpu_count()
        assert parallelization is not None, "Cannot automatically determine CPU count!"

    with open(filePath, 'rb') as file, open(
        outputFilePath, 'wb'
    ) as compressedFile, concurrent.futures.ThreadPoolExecutor(parallelization) as pool:
        results = []
        while True:
            toCompress = file.read(frameSize)
            if not toCompress:
                break
            results.append(pool.submit(_compress_zstd, toCompress))
            while len(results) >= parallelization:
                compressedData = results.pop(0).result()
                compressedFile.write(compressedData)

        while results:
            compressedFile.write(results.pop(0).result())


def get_gzip_info(fileobj: IO[bytes]) -> Optional[Tuple[str, int]]:
    id1, id2, compression, flags, mtime, _, _ = struct.unpack('<BBBBLBB', fileobj.read(10))
    if id1 != 0x1F or id2 != 0x8B or compression != 0x08:
        return None

    if flags & (1 << 2) != 0:
        fileobj.read(struct.unpack('<U', fileobj.read(2))[0])

    if flags & (1 << 3) != 0:
        name = b''
        c = fileobj.read(1)
        while c != b'\0':
            name += c
            c = fileobj.read(1)
        return name.decode(), mtime

    return None


def detect_compression(
    fileobj: IO[bytes],
    prioritizedBackends: Optional[List[str]] = None,
    printDebug: int = 0,
) -> Optional[FileFormatID]:
    # isinstance(fileobj, io.IOBase) does not work for everything, e.g., for paramiko.sftp_file.SFTPFile
    # because it does not inherit from io.IOBase. Therefore, do duck-typing and test for required methods.
    expectedMethods = ['seekable', 'seek', 'read', 'tell']
    isNotFileObject = any(not hasattr(fileobj, method) for method in expectedMethods)
    if isNotFileObject or not fileobj.seekable():
        if printDebug >= 2:
            seekable = None if isNotFileObject else fileobj.seekable()
            print(
                f"[Warning] Cannot detect compression for given Python object {fileobj} "
                f"because it does not look like a file object or is not seekable ({seekable})."
            )
        if printDebug >= 3:
            print(dir(fileobj))
            for name in ['readable', 'seekable', 'writable', 'closed', 'tell']:
                method = getattr(fileobj, name, None)
                if method is not None:
                    print(f"  fileobj.{name}:", method() if callable(method) else method)
            traceback.print_exc()
        return None

    oldOffset = fileobj.tell()
    for compressionId in COMPRESSION_FORMATS:
        if not might_be_format(fileobj, compressionId):
            continue

        backend = find_available_backend(compressionId, prioritizedBackends=prioritizedBackends)
        # If no appropriate module exists, then don't do any further checks.
        if not backend:
            if printDebug >= 1:
                print(
                    f"[Warning] A given file with magic bytes for {compressionId.name} could not be opened because "
                    "no appropriate Python module could be loaded. Are some dependencies missing? To install "
                    "ratarmountcore with all dependencies do: python3 -m pip install --user ratarmountcore[full]"
                )
            return None

        try:
            # Disable parallelization because it may lead to unnecessary expensive prefetching.
            # We only need to read 1 byte anyway to verify correctness.
            compressedFileobj = backend.open(fileobj, parallelization=1)

            # Reading 1B from a single-frame zst file might require decompressing it fully in order
            # to get uncompressed file size! Avoid that. The magic bytes should suffice mostly.
            # TODO: Make indexed_zstd not require the uncompressed size for the read call.
            if compressionId != FID.ZSTANDARD:
                compressedFileobj.read(1)
            compressedFileobj.close()
            fileobj.seek(oldOffset)
            return compressionId
        except Exception as e:
            if printDebug >= 2:
                print(f"[Warning] A given file with magic bytes for {compressionId} could not be opened because:")
                print(e)
            if printDebug >= 3:
                traceback.print_exc()
            fileobj.seek(oldOffset)

    return None


def use_rapidgzip(
    fileobj: IO[bytes],
    gzipSeekPointSpacing: int = 16 * 1024 * 1024,
    prioritizedBackends: Optional[List[str]] = None,
    printDebug: int = 0,
) -> bool:
    if fileobj is None:
        return False

    if 'rapidgzip' not in sys.modules:
        print("[Warning] Cannot use rapidgzip for access to gzip file because it is not installed. Try:")
        print("[Warning]     python3 -m pip install --user rapidgzip")
        return False

    # Check whether indexed_gzip might have a higher priority than rapidgzip if both are listed.
    if (
        prioritizedBackends
        and 'indexed_gzip' in prioritizedBackends
        and (
            (
                'rapidgzip' in prioritizedBackends
                and prioritizedBackends.index('indexed_gzip') < prioritizedBackends.index('rapidgzip')
            )
            or 'rapidgzip' not in prioritizedBackends
        )
    ):
        # Low index have higher priority (because normally the list would be checked from lowest indexes).
        return False

    # Only allow mounting of real files. Rapidgzip does work with Python file objects but we don't want to
    # mount recursive archives all with the parallel gzip decoder because then the cores would be oversubscribed!
    # Similarly, small files would result in being wholly cached into memory, which probably isn't what the user
    # had intended by using ratarmount?
    isRealFile = hasattr(fileobj, 'name') and fileobj.name and os.path.isfile(fileobj.name)
    hasMultipleChunks = isRealFile and os.stat(fileobj.name).st_size >= 4 * gzipSeekPointSpacing
    if not hasMultipleChunks:
        if printDebug >= 2:
            print("[Info] Do not reopen with rapidgzip backend because:")
            if not isRealFile:
                print("[Info]  - the file to open is a recursive file, which limits the usability of ")
                print("[Info]    parallel decompression.")
            if not hasMultipleChunks:
                print("[Info]  - is too small to qualify for parallel decompression.")
        return False

    return True


def open_compressed_file(
    fileobj: IO[bytes],
    gzipSeekPointSpacing: int,
    parallelizations: Dict[str, int],
    enabledBackends: Optional[List[str]] = None,
    prioritizedBackends: Optional[List[str]] = None,
    printDebug: int = 0,
) -> Tuple[Any, Optional[IO[bytes]], Optional[FileFormatID]]:
    """
    Opens a file possibly undoing the compression.
    Returns (tar_file_obj, raw_file_obj, compression).
    raw_file_obj will be none if compression is None.
    """
    compression = detect_compression(fileobj, prioritizedBackends=prioritizedBackends, printDebug=printDebug)
    if printDebug >= 3:
        print(f"[Info] Detected compression {compression} for file object:", fileobj)
    if not compression:
        return fileobj, None, compression

    matchingBackends = [
        backend
        for backend, info in COMPRESSION_BACKENDS.items()
        if (enabledBackends is None or backend in enabledBackends) and compression in info.formats
    ]
    if not matchingBackends:
        return fileobj, None, compression

    backend = find_available_backend(
        compression, enabledBackends=enabledBackends, prioritizedBackends=prioritizedBackends
    )
    if not backend:
        packages = [
            package for backend in matchingBackends for _, package in COMPRESSION_BACKENDS[backend].requiredModules
        ]
        raise CompressionError(
            f"Cannot open a {compression} compressed TAR file '{fileobj.name}' "
            f"without any of these packages: {packages}"
        )

    parallelization = 1

    if compression == FID.GZIP:
        if use_rapidgzip(
            fileobj,
            gzipSeekPointSpacing=gzipSeekPointSpacing,
            prioritizedBackends=prioritizedBackends,
            printDebug=printDebug,
        ):
            isRealFile = hasattr(fileobj, 'name') and fileobj.name and os.path.isfile(fileobj.name)
            parallelization = (
                1
                if isRealFile and is_on_slow_drive(fileobj.name)
                else parallelizations.get(
                    'rapidgzip-gzip', parallelizations.get('rapidgzip', parallelizations.get('', 1))
                )
            )
            if printDebug >= 3:
                print(
                    f"[Info] Parallelization to use for rapidgzip backend: {parallelization}, "
                    f"slow drive detected: {is_on_slow_drive(fileobj.name)}"
                )
            decompressedFileObject = rapidgzip.RapidgzipFile(
                fileobj, parallelization=parallelization, verbose=printDebug >= 2, chunk_size=gzipSeekPointSpacing
            )
        else:
            # The buffer size must be much larger than the spacing or else there will be large performance penalties
            # even for reading sequentially, see https://github.com/pauldmccarthy/indexed_gzip/issues/89
            # Use 4x spacing because each raw read seeks from the last index point even if the position did not
            # change since the last read call. On average, this incurs an overhead of spacing / 2. For 3x spacing,
            # thisoverhead would be 1/6 = 17%, which should be negligible. The increased memory-usage is not an
            # issue because internally many buffers are allocated with 4 * spacing size.
            bufferSize = max(3 * 1024 * 1024, 3 * gzipSeekPointSpacing)
            # drop_handles keeps a file handle opening as is required to call tell() during decoding
            decompressedFileObject = indexed_gzip.IndexedGzipFile(
                fileobj=fileobj, drop_handles=False, spacing=gzipSeekPointSpacing, buffer_size=bufferSize
            )
    elif compression == FID.BZIP2:
        parallelization = parallelizations.get('rapidgzip-bzip2', parallelizations.get('', 1))
        decompressedFileObject = rapidgzip.IndexedBzip2File(fileobj, parallelization=parallelization)  # type: ignore
    elif (
        compression == FID.XZ
        and xz
        and parallelizations.get('xz', parallelizations.get('', 1)) != 1
        and hasattr(fileobj, 'name')
        and os.path.isfile(fileobj.name)
        and platform.system() == 'Linux'
    ):
        block_boundaries = getattr(fileobj, 'block_boundaries', [])
        decompressedFileObject = backend.open(fileobj)
        if block_boundaries and len(block_boundaries) > 1:
            parallelization = parallelizations.get('xz', parallelizations.get('', 1))
            decompressedFileObject.close()
            decompressedFileObject = ParallelXZReader(fileobj.name, parallelization=parallelization)
    else:
        decompressedFileObject = backend.open(fileobj)

    if printDebug >= 3:
        print(
            f"[Info] Undid {compression} file compression by using: {type(decompressedFileObject).__name__} "
            f"with parallelization={parallelization}"
        )

    return decompressedFileObject, fileobj, compression
